r"""
`p`-adic Base Generic

A superclass for implementations of `\ZZ_p` and `\QQ_p`.

AUTHORS:

- David Roe
"""

#*****************************************************************************
#       Copyright (C) 2007-2013 David Roe <roed.math@gmail.com>
#                               William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

from .padic_generic import pAdicGeneric
from .misc import precprint
from sage.rings.padics.pow_computer import PowComputer
from sage.rings.padics.padic_capped_relative_element import pAdicCoercion_ZZ_CR, pAdicCoercion_QQ_CR, pAdicConvert_QQ_CR
from sage.rings.padics.padic_capped_absolute_element import pAdicCoercion_ZZ_CA, pAdicConvert_QQ_CA
from sage.rings.padics.padic_fixed_mod_element import pAdicCoercion_ZZ_FM, pAdicConvert_QQ_FM
from sage.rings.padics.padic_floating_point_element import pAdicCoercion_ZZ_FP, pAdicCoercion_QQ_FP, pAdicConvert_QQ_FP


class pAdicBaseGeneric(pAdicGeneric):
    _implementation = 'GMP'

    def __init__(self, p, prec, print_mode, names, element_class, category=None):
        """
        Initialization.

        TESTS::

            sage: R = Zp(5)  # indirect doctest
        """
        if self.is_relaxed():
            from sage.rings.padics.pow_computer_flint import PowComputer_flint
            self.prime_pow = PowComputer_flint(p, 1, 1, 1, self.is_field())
        else:
            self.prime_pow = PowComputer(p, max(min(prec - 1, 30), 1), prec, self.is_field(), self._prec_type())
        pAdicGeneric.__init__(self, self, p, prec, print_mode, names, element_class, category=category)
        if self.is_field():
            if self.is_capped_relative():
                coerce_list = [pAdicCoercion_ZZ_CR(self), pAdicCoercion_QQ_CR(self)]
                convert_list = []
            elif self.is_floating_point():
                coerce_list = [pAdicCoercion_ZZ_FP(self), pAdicCoercion_QQ_FP(self)]
                convert_list = []
            elif self.is_lattice_prec():
                coerce_list = [QQ]
                convert_list = []
            elif self.is_relaxed():
                coerce_list = [QQ]
                convert_list = []
            else:
                raise RuntimeError
        elif self.is_capped_relative():
            coerce_list = [pAdicCoercion_ZZ_CR(self)]
            convert_list = [pAdicConvert_QQ_CR(self)]
        elif self.is_capped_absolute():
            coerce_list = [pAdicCoercion_ZZ_CA(self)]
            convert_list = [pAdicConvert_QQ_CA(self)]
        elif self.is_fixed_mod():
            coerce_list = [pAdicCoercion_ZZ_FM(self)]
            convert_list = [pAdicConvert_QQ_FM(self)]
        elif self.is_floating_point():
            coerce_list = [pAdicCoercion_ZZ_FP(self)]
            convert_list = [pAdicConvert_QQ_FP(self)]
        elif self.is_lattice_prec():
            coerce_list = [ZZ]
            convert_list = [QQ]
        elif self.is_relaxed():
            coerce_list = [ZZ]
            convert_list = [QQ]
        else:
            raise RuntimeError
        self.Element = element_class
        self._populate_coercion_lists_(coerce_list=coerce_list, convert_list=convert_list)

    def _repr_(self, do_latex=False):
        r"""
        Return a print representation of this `p`-adic ring or field.

        EXAMPLES::

            sage: K = Zp(17); K  # indirect doctest
            17-adic Ring with capped relative precision 20
            sage: latex(K)
            \Bold{Z}_{17}
            sage: K = ZpCA(17); K  # indirect doctest
            17-adic Ring with capped absolute precision 20
            sage: latex(K)
            \Bold{Z}_{17}
            sage: K = ZpFP(17); K  # indirect doctest
            17-adic Ring with floating precision 20
            sage: latex(K)
            \Bold{Z}_{17}
            sage: K = ZpFM(7); K
            7-adic Ring of fixed modulus 7^20
            sage: latex(K)  # indirect doctest
            \Bold{Z}_{7}
            sage: K = ZpLF(2); K   # indirect doctest
            doctest:...: FutureWarning: This class/method/function is marked as experimental. It, its functionality or its interface might change without a formal deprecation.
            See https://github.com/sagemath/sage/issues/23505 for details.
            2-adic Ring with lattice-float precision
            sage: latex(K)
            \Bold{Z}_{2}
            sage: K = Qp(17); K  # indirect doctest
            17-adic Field with capped relative precision 20
            sage: latex(K)
            \Bold{Q}_{17}
            sage: K = QpFP(17); K  # indirect doctest
            17-adic Field with floating precision 20
            sage: latex(K)
            \Bold{Q}_{17}
            sage: K = QpLC(2); K   # indirect doctest
            2-adic Field with lattice-cap precision
            sage: latex(K)
            \Bold{Q}_{2}
        """
        if do_latex:
            if self.is_field():
                s = r"\Bold{Q}_{%s}" % self.prime()
            else:
                s = r"\Bold{Z}_{%s}" % self.prime()
            if hasattr(self, '_label') and self._label:
                s = r"\verb'%s' (\simeq %s)" % (self._label, s)
        else:
            s = "Field " if self.is_field() else "Ring "
            s = "%s-adic " % self.prime() + s + precprint(self._prec_type(), self.precision_cap(), self.prime())
            if hasattr(self, '_label') and self._label:
                s += " (label: %s)" % self._label
        return s

    def exact_field(self):
        """
        Return the rational field.

        For compatibility with extensions of `p`-adics.

        EXAMPLES::

            sage: Zp(5).exact_field()
            Rational Field
        """
        from sage.rings.rational_field import QQ
        return QQ

    def exact_ring(self):
        """
        Return the integer ring.

        EXAMPLES::

            sage: Zp(5).exact_ring()
            Integer Ring
        """
        from sage.rings.integer_ring import ZZ
        return ZZ

    def is_isomorphic(self, ring) -> bool:
        r"""
        Return whether ``self`` and ``ring`` are isomorphic, i.e. whether
        ``ring`` is an implementation of `\ZZ_p` for the same prime as ``self``.

        INPUT:

        - ``self`` -- a `p`-adic ring

        - ``ring`` -- a ring

        OUTPUT: ``boolean`` -- whether ``ring`` is an implementation of \ZZ_p`
        for the same prime as ``self``

        EXAMPLES::

            sage: R = Zp(5, 15, print_mode='digits'); S = Zp(5, 44, print_max_terms=4); R.is_isomorphic(S)
            True
        """
        return isinstance(ring, pAdicBaseGeneric) and self.prime() == ring.prime() and self.is_field() == ring.is_field()

    def gen(self, n=0):
        """
        Return the `n`-th generator of this extension.  For base
        rings/fields, we consider the generator to be the prime.

        EXAMPLES::

            sage: R = Zp(5); R.gen()
            5 + O(5^21)
        """
        if n != 0:
            raise IndexError("only one generator")
        return self(self.prime())

    def modulus(self, exact=False):
        r"""
        Return the polynomial defining this extension.

        For compatibility with extension fields; we define the modulus to be `x-1`.

        INPUT:

        - ``exact`` -- boolean (default: ``False``); whether to return a
          polynomial with integer entries

        EXAMPLES::

            sage: Zp(5).modulus(exact=True)
            x
        """
        return self.defining_polynomial(exact=exact)

    def absolute_discriminant(self):
        """
        Return the absolute discriminant of this `p`-adic ring.

        EXAMPLES::

            sage: Zp(5).absolute_discriminant()
            1
        """
        return 1

    def discriminant(self, K=None):
        """
        Return the discriminant of this `p`-adic ring over ``K``.

        INPUT:

        - ``self`` -- a `p`-adic ring

        - ``K`` -- a sub-ring of ``self`` or ``None`` (default: ``None``)

        OUTPUT:

        integer; the discriminant of this ring over ``K`` (or the absolute
        discriminant if ``K`` is ``None``)

        EXAMPLES::

            sage: Zp(5).discriminant()
            1
        """
        if K is None or K is self:
            return 1
        raise ValueError("Ground Ring must be a subring of self")

    def is_abelian(self) -> bool:
        """
        Return whether the Galois group is abelian, i.e. ``True``.
        #should this be automorphism group?

        EXAMPLES::

            sage: R = Zp(3, 10,'fixed-mod'); R.is_abelian()
            True
        """
        return True

    def is_normal(self) -> bool:
        """
        Return whether or not this is a normal extension, i.e. ``True``.

        EXAMPLES::

            sage: R = Zp(3, 10,'fixed-mod'); R.is_normal()
            True
        """
        return True

    def uniformizer(self):
        """
        Return a uniformizer for this ring.

        EXAMPLES::

            sage: R = Zp(3,5,'fixed-mod', 'series')
            sage: R.uniformizer()
            3
        """
        return self(self.prime_pow._prime())

    def uniformizer_pow(self, n):
        """
        Return the `n`-th power of the uniformizer of ``self`` (as
        an element of ``self``).

        EXAMPLES::

            sage: R = Zp(5)
            sage: R.uniformizer_pow(5)
            5^5 + O(5^25)
            sage: R.uniformizer_pow(infinity)
            0
        """
        return self(self.prime_pow(n))

    def _uniformizer_print(self):
        """
        Return how the uniformizer is supposed to print.

        EXAMPLES::

            sage: R = Zp(5, names='pi'); R._uniformizer_print()
            'pi'
        """
        return self.variable_name()

    def has_pth_root(self):
        r"""
        Return whether or not `\ZZ_p` has a primitive `p`-th
        root of unity.

        EXAMPLES::

            sage: Zp(2).has_pth_root()
            True
            sage: Zp(17).has_pth_root()
            False
        """
        return (self.prime() == 2)

    def has_root_of_unity(self, n):
        r"""
        Return whether or not `\ZZ_p` has a primitive `n`-th
        root of unity.

        INPUT:

        - ``self`` -- a `p`-adic ring

        - ``n`` -- integer

        OUTPUT:

        boolean; whether ``self`` has primitive `n`-th root of unity

        EXAMPLES::

            sage: R=Zp(37)
            sage: R.has_root_of_unity(12)
            True
            sage: R.has_root_of_unity(11)
            False
        """
        if (self.prime() == 2):
            return n.divides(2)
        return n.divides(self.prime() - 1)

    def zeta(self, n=None):
        r"""
        Return a generator of the group of roots of unity.

        INPUT:

        - ``self`` -- a `p`-adic ring

        - ``n`` -- integer or ``None`` (default: ``None``)

        OUTPUT:

        ``element``; a generator of the `n`-th roots of unity, or a generator
        of the full group of roots of unity if ``n`` is ``None``

        EXAMPLES::

            sage: R = Zp(37,5)
            sage: R.zeta(12)
            8 + 24*37 + 37^2 + 29*37^3 + 23*37^4 + O(37^5)
        """
        if (self.prime() == 2):
            if (n is None) or (n == 2):
                return self(-1)
            if n == 1:
                return self(1)
            raise ValueError("No, %sth root of unity in self" % n)
        else:
            from sage.rings.finite_rings.finite_field_constructor import GF
            return self.teichmuller(GF(self.prime()).zeta(n).lift())

    def zeta_order(self):
        """
        Return the order of the group of roots of unity.

        EXAMPLES::

            sage: R = Zp(37); R.zeta_order()
            36
            sage: Zp(2).zeta_order()
            2
        """
        if (self.prime() == 2):
            return 2
        return self.prime() - 1

    def plot(self, max_points=2500, **args):
        r"""
        Create a visualization of this `p`-adic ring as a fractal
        similar to a generalization of the Sierpi\'nski
        triangle.

        The resulting image attempts to capture the
        algebraic and topological characteristics of `\ZZ_p`.

        INPUT:

        - ``max_points`` -- the maximum number or points to plot,
          which controls the depth of recursion (default: 2500)

        - ``**args`` -- color, size, etc. that are passed to the
          underlying point graphics objects

        REFERENCES:

        - Cuoco, A. ''Visualizing the `p`-adic Integers'', The
          American Mathematical Monthly, Vol. 98, No. 4 (Apr., 1991),
          pp. 355-364

        EXAMPLES::

            sage: Zp(3).plot()                                                          # needs sage.plot
            Graphics object consisting of 1 graphics primitive
            sage: Zp(5).plot(max_points=625)                                            # needs sage.plot
            Graphics object consisting of 1 graphics primitive
            sage: Zp(23).plot(rgbcolor=(1,0,0))                                         # needs sage.plot
            Graphics object consisting of 1 graphics primitive
        """
        if 'pointsize' not in args:
            args['pointsize'] = 1
        from sage.misc.mrange import cartesian_product_iterator
        from sage.rings.real_double import RDF
        from sage.plot.point import point as points
        p = self.prime()
        phi = 2*RDF.pi()/p
        V = RDF**2
        vs = [V([(phi*t).sin(), (phi*t).cos()]) for t in range(p)]
        all = []
        depth = max(RDF(max_points).log(p).floor(), 1)
        scale = min(RDF(1.5/p), 1/RDF(3))
        pts = [vs]*depth
        if depth == 1 and 23 < p < max_points:
            extras = int(max_points/p)
            if p/extras > 5:
                pts = [vs]*depth + [vs[::extras]]
        for digits in cartesian_product_iterator(pts):
            p = sum([v * scale**n for n, v in enumerate(digits)])
            all.append(tuple(p))
        g = points(all, **args)
        # Set default plotting options
        g.axes(False)
        g.set_aspect_ratio(1)
        return g



from sage.misc.lazy_import import lazy_import
lazy_import('sage.rings.asymptotic.asymptotic_ring', 'AsymptoticRing')
lazy_import('sage.rings.asymptotic.asymptotic_expansion_generators',
            'asymptotic_expansions')
del lazy_import



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_airy.h":

  double  gsl_sf_airy_Ai(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Ai_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Bi(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Bi_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Ai_scaled(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Ai_scaled_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Bi_scaled(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Bi_scaled_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Ai_deriv(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Ai_deriv_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Bi_deriv(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Bi_deriv_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Ai_deriv_scaled(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Ai_deriv_scaled_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_Bi_deriv_scaled(double x, gsl_mode_t mode)

  int  gsl_sf_airy_Bi_deriv_scaled_e(double x, gsl_mode_t mode, gsl_sf_result * result)

  double  gsl_sf_airy_zero_Ai(unsigned int s)

  int  gsl_sf_airy_zero_Ai_e(unsigned int s, gsl_sf_result * result)

  double  gsl_sf_airy_zero_Bi(unsigned int s)

  int  gsl_sf_airy_zero_Bi_e(unsigned int s, gsl_sf_result * result)

  double  gsl_sf_airy_zero_Ai_deriv(unsigned int s)

  int  gsl_sf_airy_zero_Ai_deriv_e(unsigned int s, gsl_sf_result * result)

  double  gsl_sf_airy_zero_Bi_deriv(unsigned int s)

  int  gsl_sf_airy_zero_Bi_deriv_e(unsigned int s, gsl_sf_result * result)




# sage.doctest: needs sage.combinat sage.rings.finite_rings
r"""
Univariate dense skew polynomials over a field with a finite order automorphism

AUTHOR::

- Xavier Caruso (2012-06-29): initial version

- Arpit Merchant (2016-08-04): improved docstrings, fixed doctests
  and refactored classes and methods
"""

# ***************************************************************************
#    Copyright (C) 2012 Xavier Caruso <xavier.caruso@normalesup.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 2 of the License, or
#    (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************
from sage.structure.parent cimport Parent
from sage.structure.element cimport RingElement
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.polynomial.skew_polynomial_element cimport SkewPolynomial_generic_dense


cdef class SkewPolynomial_finite_order_dense(SkewPolynomial_generic_dense):
    def __init__(self, parent, x=None, int check=1, int construct=0, **kwds):
        """
        This method constructs a generic dense skew polynomial over a field equipped
        with an automorphism of finite order.

        INPUT:

        - ``parent`` -- parent of ``self``

        - ``x`` -- list of coefficients from which ``self`` can be constructed

        - ``check`` -- flag variable to normalize the polynomial

        - ``construct`` -- boolean (default: ``False``)

        TESTS::

            sage: R.<t> = GF(5^3)
            sage: Frob = R.frobenius_endomorphism()
            sage: S.<x> = R['x', Frob]; S
            Ore Polynomial Ring in x over Finite Field in t of size 5^3 twisted by t |--> t^5

        We create a skew polynomial from a list::

            sage: S([t,1])
            x + t

        from another skew polynomial::

            sage: S(x^2 + t)
            x^2 + t

        from a constant::

            sage: x = S(t^2 + 1); x
            t^2 + 1
            sage: x.parent() is S
            True
        """
        SkewPolynomial_generic_dense.__init__(self, parent, x, check, construct, **kwds)
        self._norm = None
        self._charpoly = None
        self._optbound = None

    cdef _matphir_c(self):
        r"""
        Return the matrix of the multiplication by `X^r` on
        the quotient `K[X,\sigma] / K[X,\sigma]*self`.
        """
        from sage.matrix.constructor import matrix
        parent = self._parent
        cdef Py_ssize_t i, j, col, exp, n
        cdef Py_ssize_t d = self.degree()
        cdef Py_ssize_t r = parent._order
        cdef k = parent.base_ring()
        cdef RingElement zero = k(0)
        cdef RingElement one = k(1)
        cdef list line, phir = []
        if r < d:
            for i from 0 <= i < d-r:
                line = d * [zero]
                line[r+i] = one
                phir.append(line)
            col = d-r
            exp = d
        else:
            col = 0
            exp = r
        cdef SkewPolynomial_finite_order_dense powx = <SkewPolynomial_finite_order_dense>self._new_c([zero, one], parent)
        cdef SkewPolynomial_finite_order_dense v
        if (exp % 2 == 1):
            v = <SkewPolynomial_finite_order_dense>self._new_c([zero,one], parent)
            if self.degree() == 1:
                v %= self
        else:
            v = <SkewPolynomial_finite_order_dense>self._new_c([one], parent)
        exp = exp >> 1
        n = 1
        while exp != 0:
            powx = (powx.conjugate(n) * powx) % self
            n = n << 1
            if (exp % 2 == 1):
                v = v.conjugate(n)
                v = (v * powx) % self
            exp = exp >> 1
        line = v.list()
        line += (d - len(line)) * [zero]
        phir.append(line)
        for j from col+1 <= j < d:
            v <<= 1
            v = v.conjugate(1) % self
            line = v._coeffs[:] + (d - len(v._coeffs)) * [zero]
            phir.append(line)
        return matrix(k, phir)

    cdef _matmul_c(self):
        r"""
        Return the matrix of the multiplication by ``self`` on
        `K[X,\sigma]` considered as a free module over `K[X^r]`
        (here `r` is the order of `\sigma`).
        """
        from sage.matrix.constructor import matrix
        cdef Py_ssize_t i, j, deb, k, r = self.parent()._order
        cdef Py_ssize_t d = self.degree ()
        cdef Parent base_ring = <Parent?>self.parent().base_ring()
        cdef RingElement zero = <RingElement?>base_ring(0)
        cdef Polk = PolynomialRing (base_ring, 'xr')
        cdef list M = []
        cdef list l = self.list()
        for j in range(r):
            for i in range(r):
                if i < j:
                    pol = [zero]
                    deb = i-j+r
                else:
                    pol = []
                    deb = i-j
                for k from deb <= k <= d by r:
                    pol.append(l[k])
                M.append(Polk(pol))
            for i from 0 <= i <= d:
                l[i] = self._parent.twisting_morphism()(l[i])
        return matrix(Polk, r, r, M)

    def reduced_trace(self, var=None):
        r"""
        Return the reduced trace of this skew polynomial.

        INPUT:

        - ``var`` -- string or ``False`` or ``None`` (default: ``None``);
          the variable name. If ``False``, return the list of coefficients.

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<x> = k['x', Frob]
            sage: a = x^3 + (2*t^2 + 3)*x^2 + (4*t^2 + t + 4)*x + 2*t^2 + 2
            sage: tr = a.reduced_trace(); tr
            3*z + 4

        The reduced trace lies in the center of `S`, which is a univariate
        polynomial ring in the variable `z = x^3` over `\GF{5}`::

            sage: tr.parent()
            Univariate Polynomial Ring in z over Finite Field of size 5
            sage: tr.parent() is S.center()
            True

        We can use explicit conversion to view ``tr`` as a skew polynomial::

            sage: S(tr)
            3*x^3 + 4

        By default, the name of the central variable is usually ``z`` (see
        :meth:`~sage.rings.polynomial.skew_polynomial_ring.SkewPolynomialRing_finite_order.center`
        for more details about this).
        However, the user can specify a different variable name if desired::

            sage: a.reduced_trace(var='u')
            3*u + 4

        When passing in ``var=False``, a tuple of coefficients (instead of
        an actual polynomial) is returned::

            sage: a.reduced_trace(var=False)
            (4, 3)

        TESTS:

        We check that the reduced trace is additive::

            sage: a = S.random_element(degree=5)
            sage: b = S.random_element(degree=7)
            sage: a.reduced_trace() + b.reduced_trace() == (a+b).reduced_trace()
            True

        .. SEEALSO::

            :meth:`reduced_norm`, :meth:`reduced_charpoly`
        """
        order = self.parent()._order
        twisting_morphism = self.parent().twisting_morphism()
        coeffs = []
        for i in range(0, self.degree()+1, order):
            tr = c = self._coeffs[i]
            for _ in range(order-1):
                tr = c + twisting_morphism(tr)
            coeffs.append(tr)
        if var is False:
            return tuple(coeffs)
        Z = self.parent().center(name=var)
        return Z(coeffs)

    def reduced_norm(self, var=None):
        r"""
        Return the reduced norm of this skew polynomial.

        INPUT:

        - ``var`` -- string or ``False`` or ``None`` (default: ``None``);
          the variable name. If ``False``, return the list of coefficients.

        .. NOTE::

            The result is cached.

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<x> = k['x', Frob]
            sage: a = x^3 + (2*t^2 + 3)*x^2 + (4*t^2 + t + 4)*x + 2*t^2 + 2
            sage: N = a.reduced_norm(); N
            z^3 + 4*z^2 + 4

        The reduced norm lies in the center of `S`, which is a univariate
        polynomial ring in the variable `z = x^3` over `\GF{5}`::

            sage: N.parent()
            Univariate Polynomial Ring in z over Finite Field of size 5
            sage: N.parent() is S.center()
            True

        We can use explicit conversion to view ``N`` as a skew polynomial::

            sage: S(N)
            x^9 + 4*x^6 + 4

        By default, the name of the central variable is usually ``z`` (see
        :meth:`~sage.rings.polynomial.skew_polynomial_ring.SkewPolynomialRing_finite_order.center`
        for more details about this).
        However, the user can specify a different variable name if desired::

            sage: a.reduced_norm(var='u')
            u^3 + 4*u^2 + 4

        When passing in ``var=False``, a tuple of coefficients (instead of
        an actual polynomial) is returned::

            sage: a.reduced_norm(var=False)
            (4, 0, 4, 1)

        TESTS:

        We check that `N` is a multiple of `a`::

            sage: S(N).is_right_divisible_by(a)
            True
            sage: S(N).is_left_divisible_by(a)
            True

        We check that the reduced norm is a multiplicative map::

            sage: a = S.random_element(degree=5)
            sage: b = S.random_element(degree=7)
            sage: a.reduced_norm() * b.reduced_norm() == (a*b).reduced_norm()
            True

        We check that the reduced norm is correctly computed for a
        constant polynomial::

            sage: c = k.random_element()
            sage: S(c).reduced_norm() == c.norm()
            True

        ALGORITHM:

        If `r` (= the order of the twist map) is small compared
        to `d` (= the degree of this skew polynomial), the reduced
        norm is computed as the determinant of the multiplication
        by `P` (= this skew polynomial) acting on `K[X,\sigma]`
        (= the underlying skew ring) viewed as a free module of
        rank `r` over `K[X^r]`.

        Otherwise, the reduced norm is computed as the characteristic
        polynomial of the left multiplication by `X` on the quotient
        `K[X,\sigma] / K[X,\sigma] P` (which is a `K`-vector space
        of dimension `d`).

        .. SEEALSO::

            :meth:`reduced_trace`, :meth:`reduced_charpoly`
        """
        if self._norm is None:
            if self.is_zero():
                self._norm = 0
            else:
                parent = self._parent
                section = parent._embed_constants.section()
                exp = (parent.base_ring().cardinality() - 1) / (parent._constants.cardinality() - 1)
                order = self.parent()._order
                lc = section(self.leading_coefficient()**exp)
                if self.degree() == 0:
                    self._norm = (lc,)
                elif order < self.degree():
                    M = self._matmul_c()
                    self._norm = tuple(lc*section(x) for x in M.determinant().monic().list())
                else:
                    charpoly = self._matphir_c().characteristic_polynomial()
                    self._norm = tuple(lc*section(x) for x in charpoly.list())
        if var is False:
            return self._norm
        center = self.parent().center(name=var)
        return center(self._norm)

    def reduced_charpoly(self, var=None):
        r"""
        Return the reduced characteristic polynomial of this
        skew polynomial.

        INPUT:

        - ``var`` -- string, a pair of strings or ``None``
          (default: ``None``); the variable names used for the
          characteristic polynomial and the center

        .. NOTE::

            The result is cached.

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<u> = k['u', Frob]
            sage: a = u^3 + (2*t^2 + 3)*u^2 + (4*t^2 + t + 4)*u + 2*t^2 + 2
            sage: chi = a.reduced_charpoly()
            sage: chi
            x^3 + (2*z + 1)*x^2 + (3*z^2 + 4*z)*x + 4*z^3 + z^2 + 1

        The reduced characteristic polynomial has coefficients in the center
        of `S`, which is itself a univariate polynomial ring in the variable
        `z = u^3` over `\GF{5}`. Hence it appears as a bivariate polynomial::

            sage: chi.parent()
            Univariate Polynomial Ring in x over Univariate Polynomial Ring in z over Finite Field of size 5

        The constant coefficient of the reduced characteristic polynomial is
        the reduced norm, up to a sign::

            sage: chi[0] == -a.reduced_norm()
            True

        Its coefficient of degree `\deg(a) - 1` is the opposite of the reduced
        trace::

            sage: chi[2] == -a.reduced_trace()
            True

        By default, the name of the variable of the reduced characteristic
        polynomial is ``x`` and the name of central variable is usually ``z``
        (see :meth:`~sage.rings.polynomial.skew_polynomial_ring.SkewPolynomialRing_finite_order.center`
        for more details about this).
        The user can specify different names if desired::

            sage: a.reduced_charpoly(var='T')  # variable name for the characteristic polynomial
            T^3 + (2*z + 1)*T^2 + (3*z^2 + 4*z)*T + 4*z^3 + z^2 + 1

            sage: a.reduced_charpoly(var=('T', 'c'))
            T^3 + (2*c + 1)*T^2 + (3*c^2 + 4*c)*T + 4*c^3 + c^2 + 1

        .. SEEALSO::

            :meth:`reduced_trace`, :meth:`reduced_norm`

        TESTS:

        We test that the cache works correctly (see :issue:`39883`)::

            sage: f = u^5 + t*u^4 + t^2*u^3 + t^3*u^2 + t^4*u + t^5
            sage: f.reduced_norm()
            z^5 + 2*z^4 + 4*z^3 + z^2 + 4*z + 2
            sage: f.reduced_charpoly()
            x^3 + z*x^2 + (2*z + 3)*x + 4*z^5 + 3*z^4 + z^3 + 4*z^2 + z + 3
            sage: f.reduced_norm()
            z^5 + 2*z^4 + 4*z^3 + z^2 + 4*z + 2
        """
        if self._charpoly is None:
            M = self._matmul_c()
            chi = M.charpoly()
            self._charpoly = [tuple(c.list()) for c in chi.list()]
            if self._norm is None:
                if len(self._charpoly) % 2:
                    self._norm = self._charpoly[0]
                else:
                    self._norm = tuple((-chi[0]).list())
        varcenter = None
        if var is None:
            varcharpoly = 'x'
        elif isinstance(var, (tuple, list)) and len(var) == 2:
            varcharpoly, varcenter = var
        else:
            varcharpoly = var
        center = self.parent().center(name=varcenter)
        coeffs = [center(c) for c in self._charpoly]
        return PolynomialRing(center, name=varcharpoly)(coeffs)

    def is_central(self) -> bool:
        r"""
        Return ``True`` if this skew polynomial lies in the center.

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<x> = k['x', Frob]

            sage: x.is_central()
            False
            sage: (t*x^3).is_central()
            False
            sage: (x^6 + x^3).is_central()
            True
        """
        center = self.parent()._working_center
        try:
            center(self)
            return True
        except ValueError:
            return False

    def bound(self):
        r"""
        Return a bound of this skew polynomial (i.e. a multiple
        of this skew polynomial lying in the center).

        .. NOTE::

            Since `b` is central, it divides a skew polynomial
            on the left iff it divides it on the right

        ALGORITHM:

        #. Sage first checks whether self is itself in the
           center. It if is, it returns ``self``

        #. If an optimal bound was previously computed and
           cached, Sage returns it

        #. Otherwise, Sage returns the reduced norm of ``self``

        As a consequence, the output of this function may depend
        on previous computations (an example is given below).

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<x> = k['x', Frob]
            sage: Z = S.center(); Z
            Univariate Polynomial Ring in z over Finite Field of size 5

            sage: a = x^2 + (4*t + 2)*x + 4*t^2 + 3
            sage: b = a.bound(); b
            z^2 + z + 4

        We observe that the bound is explicitly given as an element of the
        center (which is a univariate polynomial ring in the variable `z`).
        We can use conversion to send it in the skew polynomial ring::

            sage: S(b)
            x^6 + x^3 + 4

        We check that `b` is divisible by `a`::

            sage: S(b).is_right_divisible_by(a)
            True
            sage: S(b).is_left_divisible_by(a)
            True

        Actually, `b` is the reduced norm of `a`::

            sage: b == a.reduced_norm()
            True

        Now, we compute the optimal bound of `a` and see that
        it affects the behaviour of ``bound()``::

            sage: a.optimal_bound()
            z + 3
            sage: a.bound()
            z + 3

        TESTS:

        We check that when the input skew polynomial lies in
        the center, the output is the skew polynomial itself::

            sage: a = Z.random_element(degree=4)
            sage: b = S(a).bound()
            sage: a == b
            True
        """
        center = self.parent().center()
        try:
            return center(self)
        except ValueError:
            pass
        if self._optbound is not None:
            return center(self._optbound)
        return self.reduced_norm()

    def optimal_bound(self):
        r"""
        Return the optimal bound of this skew polynomial (i.e.
        the monic multiple of this skew polynomial of minimal
        degree lying in the center).

        .. NOTE::

            The result is cached.

        EXAMPLES::

            sage: k.<t> = GF(5^3)
            sage: Frob = k.frobenius_endomorphism()
            sage: S.<x> = k['x', Frob]
            sage: Z = S.center(); Z
            Univariate Polynomial Ring in z over Finite Field of size 5

            sage: a = x^2 + (4*t + 2)*x + 4*t^2 + 3
            sage: b = a.optimal_bound(); b
            z + 3

        We observe that the bound is explicitly given as an element of the
        center (which is a univariate polynomial ring in the variable `z`).
        We can use conversion to send it in the skew polynomial ring::

            sage: S(b)
            x^3 + 3

        We check that `b` is divisible by `a`::

            sage: S(b).is_right_divisible_by(a)
            True
            sage: S(b).is_left_divisible_by(a)
            True
        """
        center = self.parent().center()
        if self._optbound is None:
            try:
                self._optbound = center(self).monic()
            except ValueError:
                bound = self._matphir_c().minimal_polynomial()
                section = self._parent._embed_constants.section()
                self._optbound = [section(x) for x in bound.list()]
        return center(self._optbound)

     # TODO:
     # fast multiplication
     # reduced characteristic polynomial



r"""
Galois representations attached to elliptic curves

Given an elliptic curve `E` over `\QQ`
and a rational prime number `p`, the `p^n`-torsion
`E[p^n]` points of `E` is a representation of the
absolute Galois group `G_{\QQ}` of `\QQ`. As `n` varies
we obtain the Tate module `T_p E` which is a
a representation of `G_\QQ` on a free `\ZZ_p`-module
of rank `2`. As `p` varies the representations
are compatible.

Currently sage can decide whether the Galois module
`E[p]` is reducible, i.e., if `E` admits an isogeny
of degree `p`, and whether the image of
the representation on `E[p]` is surjective onto
`\text{Aut}(E[p]) = GL_2(\GF{p})`.

The following are the most useful functions for the class ``GaloisRepresentation``.

For the reducibility:

- ``is_reducible(p)``
- ``is_irreducible(p)``
- ``reducible_primes()``

For the image:

- ``is_surjective(p)``
- ``non_surjective()``
- ``image_type(p)``

For the classification of the representation

- ``is_semistable(p)``
- ``is_unramified(p, ell)``
- ``is_crystalline(p)``

EXAMPLES::

    sage: E = EllipticCurve('196a1')
    sage: rho = E.galois_representation()
    sage: rho.is_irreducible(7)
    True
    sage: rho.is_reducible(3)
    True
    sage: rho.is_irreducible(2)
    True
    sage: rho.is_surjective(2)
    False
    sage: rho.is_surjective(3)
    False
    sage: rho.is_surjective(5)
    True
    sage: rho.reducible_primes()
    [3]
    sage: rho.non_surjective()
    [2, 3]
    sage: rho.image_type(2)
    'The image is cyclic of order 3.'
    sage: rho.image_type(3)
    'The image is contained in a Borel subgroup as there is a 3-isogeny.'
    sage: rho.image_type(5)
    'The image is all of GL_2(F_5).'

For semi-stable curve it is known that the representation is
surjective if and only if it is irreducible::

    sage: E = EllipticCurve('11a1')
    sage: rho = E.galois_representation()
    sage: rho.non_surjective()
    [5]
    sage: rho.reducible_primes()
    [5]

For cm curves it is not true that there are only finitely many primes for which the
Galois representation mod p is surjective onto `GL_2(\GF{p})`::

    sage: E = EllipticCurve('27a1')
    sage: rho = E.galois_representation()
    sage: rho.non_surjective()
    [0]
    sage: rho.reducible_primes()
    [3]
    sage: E.has_cm()
    True
    sage: rho.image_type(11)
    'The image is contained in the normalizer of a non-split Cartan group. (cm)'

REFERENCES:

- [Ser1972]_
- [Ser1987]_
- [Coj2005]_

AUTHORS:

- chris wuthrich (02/10): moved from ell_rational_field.py.
"""

######################################################################
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
######################################################################

from math import sqrt

from sage.structure.sage_object import SageObject
import sage.arith.all as arith
from sage.rings.fast_arith import prime_range
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod as mul
from sage.misc.verbose import verbose
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.real_mpfr import RealField
from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF

lazy_import('sage.libs.pari', 'pari')


def _ex_set(p):
    """
    Give the set of the only values of trace^2/det
    that appear in a exceptional subgroup in PGL_2(F_p).

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.gal_reps import _ex_set
        sage: for p in prime_range(3,30): print("{} {}".format(p, _ex_set(p)))
        3 [0, 1, 2, 1]
        5 [0, 1, 2, 4]
        7 [0, 1, 2, 4]
        11 [0, 1, 2, 4, 9, 5]
        13 [0, 1, 2, 4]
        17 [0, 1, 2, 4]
        19 [0, 1, 2, 4, 16, 6]
        23 [0, 1, 2, 4]
        29 [0, 1, 2, 4, 25, 7]
    """
    k = GF(p)
    res = [ k(0), k(1), k(2), k(4) ]
    R = k['X']
    f = R([1,-3,1]) #(X**2 - 3*X+1)
    ro = f.roots()
    for a in ro:
        if a[0] not in res:
            res.append(a[0])
    return res


class GaloisRepresentation(SageObject):
    r"""
    The compatible family of Galois representation
    attached to an elliptic curve over the rational numbers.

    Given an elliptic curve `E` over `\QQ` and a rational
    prime number `p`, the `p^n`-torsion `E[p^n]` points of
    `E` is a representation of the absolute Galois group.
    As `n` varies we obtain the Tate module `T_p E` which is
    a representation of the absolute Galois group on a free
    `\ZZ_p`-module of rank `2`. As `p` varies the
    representations are compatible.

    EXAMPLES::

        sage: rho = EllipticCurve('11a1').galois_representation()
        sage: rho
        Compatible family of Galois representations associated to the Elliptic Curve defined by y^2 + y = x^3 - x^2 - 10*x - 20 over Rational Field
    """

    def __init__(self, E):
        r"""
        See ``GaloisRepresentation`` for documentation.

        EXAMPLES::

            sage: rho = EllipticCurve('11a1').galois_representation()
            sage: loads(rho.dumps()) == rho
            True
        """
        self.__image_type = {}
        self._E = E

    def __repr__(self):
        r"""
        String representation of the class.

        EXAMPLES::

            sage: rho = EllipticCurve([0,1]).galois_representation()
            sage: rho
            Compatible family of Galois representations associated to the Elliptic Curve defined by y^2 = x^3 + 1 over Rational Field
        """
        return "Compatible family of Galois representations associated to the " + repr(self._E)

    def __eq__(self, other):
        r"""
        Compare two Galois representations.
        We define tho compatible families of representations
        attached to elliptic curves to be isomorphic if the curves are equal.

        EXAMPLES::

            sage: rho = EllipticCurve('11a1').galois_representation()
            sage: rho2 = EllipticCurve('11a2').galois_representation()
            sage: rho == rho
            True
            sage: rho == rho2
            False
            sage: rho == 34
            False
        """
        # if rho_E = rho_E' then the L-functions agree,
        # so E and E' are isogenous
        # except for p=2
        # the mod p representations will be different
        # for p dividing the degree of the isogeny
        # anyway, there should not be a _compatible_
        # isomorphism between rho and rho' unless E
        # is isomorphic to E'
        # Note that rho can not depend on the Weierstrass model
        if type(self) is not type(other):
            return False
        return self._E.is_isomorphic(other._E)

    def elliptic_curve(self):
        r"""
        The elliptic curve associated to this representation.

        EXAMPLES::

            sage: E = EllipticCurve('11a1')
            sage: rho = E.galois_representation()
            sage: rho.elliptic_curve() == E
            True
        """
        from copy import copy
        return copy(self._E)

#####################################################################
# reducibility
#####################################################################

    def is_reducible(self, p):
        r"""
        Return ``True`` if the mod-p representation is
        reducible. This is equivalent to the existence of an
        isogeny defined over `\QQ` of degree `p` from the
        elliptic curve.

        INPUT:

        - ``p`` -- a prime number

        OUTPUT: boolean

        The answer is cached.

        EXAMPLES::

            sage: rho = EllipticCurve('121a').galois_representation()
            sage: rho.is_reducible(7)
            False
            sage: rho.is_reducible(11)
            True
            sage: EllipticCurve('11a').galois_representation().is_reducible(5)
            True
            sage: rho = EllipticCurve('11a2').galois_representation()
            sage: rho.is_reducible(5)
            True
            sage: EllipticCurve('11a2').torsion_order()
            1
        """
        try:
            return self.__is_reducible[p]
        except AttributeError:
            self.__is_reducible = {}
        except KeyError:
            pass

        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        # we do is_surjective first, since this is
        # much easier than computing isogeny_class
        t = self._is_surjective(p, A=-1)
        if t:
            self.__is_reducible[p] = False
            return False  # definitely not reducible
        isogeny_matrix = self._E.isogeny_class().matrix(fill=True)
        v = isogeny_matrix.row(0) # first row
        for a in v:
            if a != 0 and a % p == 0:
                self.__is_reducible[p] = True
                return True
        self.__is_reducible[p] = False
        return False

    def is_irreducible(self, p):
        r"""
        Return ``True`` if the mod p representation is irreducible.

        INPUT:

        - ``p`` -- a prime number

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('37b').galois_representation()
            sage: rho.is_irreducible(2)
            True
            sage: rho.is_irreducible(3)
            False
            sage: rho.is_reducible(2)
            False
            sage: rho.is_reducible(3)
            True
        """
        return not self.is_reducible(p)

    def reducible_primes(self):
        r"""
        Return a list of the primes `p` such that the mod-`p`
        representation is reducible. For all other primes the
        representation is irreducible.

        EXAMPLES::

            sage: rho = EllipticCurve('225a').galois_representation()
            sage: rho.reducible_primes()
            [3]
        """
        try:
            return self.__reducible_primes
        except AttributeError:
            pass

        E = self._E
        j = E.j_invariant()
        from .isogeny_small_degree import sporadic_j
        if j in sporadic_j: # includes all CM j-invariants
            R = [sporadic_j[j]]
        else:
            R = [l for l in [2,3,5,7,13] if len(E.isogenies_prime_degree(l)) > 0]
        self.__reducible_primes = R
        return R

#####################################################################
# image
#####################################################################

    def is_surjective(self, p, A=1000):
        r"""
        Return ``True`` if the mod-p representation is
        surjective onto `Aut(E[p]) = GL_2(\GF{p})`.

        ``False`` if it is not, or ``None`` if we were unable to
        determine whether it is or not.

        INPUT:

        - ``p`` -- integer; a prime number

        - ``A`` -- integer; a bound on the number of `a_p` to use

        OUTPUT: boolean; ``True`` if the mod-p representation is surjective
        and ``False`` if not

        The answer is cached.

        EXAMPLES::

            sage: rho = EllipticCurve('37b').galois_representation()
            sage: rho.is_surjective(2)
            True
            sage: rho.is_surjective(3)
            False

        ::

            sage: rho = EllipticCurve('121a1').galois_representation()
            sage: rho.non_surjective()
            [11]
            sage: rho.is_surjective(5)
            True
            sage: rho.is_surjective(11)
            False

            sage: rho = EllipticCurve('121d1').galois_representation()
            sage: rho.is_surjective(5)
            False
            sage: rho.is_surjective(11)
            True

        Here is a case, in which the algorithm does not return an answer::

            sage: rho = EllipticCurve([0,0,1,2580,549326]).galois_representation()
            sage: rho.is_surjective(7)

        In these cases, one can use image_type to get more information about the image::

            sage: rho.image_type(7)
            'The image is contained in the normalizer of a split Cartan group.'

        REMARKS:

        1. If `p \geq 5` then the mod-p representation is
           surjective if and only if the `p`-adic representation is
           surjective. When `p = 2, 3` there are
           counterexamples. See papers of Dokchitsers and Elkies
           for more details.

        2. For the primes `p=2` and 3, this will always answer either
           ``True`` or ``False``. For larger primes it might give ``None``.
        """
        if not arith.is_prime(p):
            raise TypeError("p (=%s) must be prime." % p)
        A = int(A)
        key = (p, A)
        try:
            return self.__is_surjective[key]
        except KeyError:
            pass
        except AttributeError:
            self.__is_surjective = {}

        ans = self._is_surjective(p, A)
        self.__is_surjective[key] = ans
        return ans

    def _is_surjective(self, p, A):
        r"""
        Helper function for ``is_surjective``.

        The value of `A` is as before, except that
        `A=-1` is a special code to stop before
        testing reducibility (to avoid an infinite loop).

        EXAMPLES::

            sage: rho = EllipticCurve('37b').galois_representation()
            sage: rho._is_surjective(7,100)
            True

        TEST:

        Test for :issue:`8451`::

            sage: E = EllipticCurve('648a1')
            sage: rho = E.galois_representation()
            sage: rho._is_surjective(5,1000)
        """
        T = self._E.torsion_subgroup().order()
        if T % p == 0 and p != 2:
            # we could probably determine the group structure directly
            self.__image_type[p] = "The image is meta-cyclic inside a Borel subgroup as there is a %s-torsion point on the curve." % p
            return False

        R = PolynomialRing(self._E.base_ring(), 'x')
        x = R.gen()

        if p == 2:
            # E is isomorphic to  [0,b2,0,8*b4,16*b6]
            b2,b4,b6,b8 = self._E.b_invariants()
            f = x**3 + b2*x**2 + 8*b4*x + 16*b6
            if not f.is_irreducible():
                if len(f.roots()) > 2:
                    self.__image_type[p] = "The image is trivial as all 2-torsion points are rational."
                else:
                    self.__image_type[p] = "The image is cyclic of order 2 as there is exactly one rational 2-torsion point."
                return False    #, '2-torsion'
            if arith.is_square(f.discriminant()):
                self.__image_type[p] = "The image is cyclic of order 3."
                return False    #, "A3"
            self.__image_type[p] = "The image is all of GL_2(F_2), i.e. a symmetric group of order 6."
            return True  #, None

        if p == 3:
            # Algorithm: Let f be the 3-division polynomial, which is
            # a polynomial of degree 4.  Then I claim that this
            # polynomial has Galois group S_4 if and only if the
            # representation rhobar_{E,3} is surjective.  If the group
            # is S_4, then S_4 is a quotient of the image of
            # rhobar_{E,3}.  Since S_4 has order 24 and GL_2(F_3)
            # has order 48, the only possibility we have to consider
            # is that the image of rhobar is isomorphic to S_4.
            # But this is not the case because S_4 is not a subgroup
            # of GL_2(F_3).    If it were, it would be normal, since
            # it would have index 2.  But there is a *unique* normal
            # subgroup of GL_2(F_3) of index 2, namely SL_2(F_3),
            # and SL_2(F_3) is not isomorphic to S_4 (S_4 has a normal
            # subgroup of index 2 and SL_2(F_3) does not.)
            # (What's a simple way to see that SL_2(F_3) is the
            # unique index-2 normal subgroup?  I didn't see an obvious
            # reason, so just used the NormalSubgroups command in MAGMA
            # and it output exactly one of index 2.)

            #sage: G = SymmetricGroup(4)
            #sage: [H.group_id() for H in G.conjugacy_classes_subgroups()]
            #[[1, 1], [2, 1], [2, 1], [3, 1], [4, 2], [4, 2], [4, 1], [6, 1], [8, 3], [12, 3], [24, 12]]
            #sage: G = GL(2,GF(3)).as_matrix_group().as_permutation_group()
            #sage: [H.group_id() for H in G.conjugacy_classes_subgroups()]
            #[[1, 1], [2, 1], [2, 1], [3, 1], [4, 2], [4, 1], [6, 2], [6, 1], [6, 1], [8, 4], [8, 1], [8, 3], [12, 4], [16, 8], [24, 3], [48, 29]]

            # Here's Noam Elkies proof for the other direction:

            #> Let E be an elliptic curve over Q.  Is the mod-3
            #> representation E[3]  surjective if and only if the
            #> (degree 4) division polynomial has Galois group S_4?  I
            #> can see why the group being S_4 implies the
            #> representation is surjective, but the converse is not
            #> clear to me.
            # I would have thought that this is the easier part: to
            # say that E[3] is surjective is to say the 3-torsion
            # field Q(E[3]) has Galois group GL_2(Z/3) over Q.  Let
            # E[3]+ be the subfield fixed by the element -1 of
            # GL_2(Z/3).  Then E[3] has Galois group PGL_2(Z/3), which
            # is identified with S_4 by its action on the four
            # 3-element subgroups of E[3].  Each such subgroup is in
            # turn determined by the x-coordinate shared by its two
            # nonzero points.  So, if E[3] is surjective then any
            # permutation of those x-coordinates is realized by some
            # element of Gal(E[3]+/Q).  Thus the Galois group of the
            # division polynomial (whose roots are those
            # x-coordinates) maps surjectively to S_4, which means it
            # equals S_4.

            f = self._E.division_polynomial(3)
            if not f.is_irreducible():
                return False   #, "reducible_3-divpoly"
            n = pari(f).polgalois()[0]
            if n == 24:
                self.__image_type[p] = "The image is all of GL_2(F_3)."
                return True   #, None
            return False   #, "3-divpoly_galgroup_order_%s"%n

        if self._E.has_cm():
            return False   #, "CM"

        # Now we try to prove that the rep IS surjective.

        Np = self._E.conductor() * p
        signs = []
        # this follows Proposition 19 in Serre.
        # there was a bug in the original implementation,
        # counter-examples were 324b1 and 324d1, 648a1 and 648c1 for p=5
        exclude_exceptional_image = False
        ex_setp = _ex_set(p)
        ell = 4
        k = GF(p)

        if A == -1:
            Am = 1000
        else:
            Am = A
        while ell < Am:
            ell = arith.next_prime(ell)
            if Np % ell != 0:
                a_ell = self._E.ap(ell)
                if a_ell % p != 0:
                    if not exclude_exceptional_image:
                        u = k(a_ell)**2 * k(ell)**(-1)
                        if u not in ex_setp:
                            exclude_exceptional_image = True
                    s = arith.kronecker(a_ell**2 - 4*ell, p)
                    if s != 0 and s not in signs:
                        signs.append(s)
                    if len(signs) == 2 and exclude_exceptional_image:
                        self.__image_type[p] = "The image is all of GL_2(F_%s)." % p
                        return True   #,None

        if A == -1: # we came in from is reducible. Now go out with False
            return False

        if self.is_reducible(p):
            return False  #, Borel

        # if we reach this, then we do not know if it is surjective. Most likely
        # not but we can't be certain. See trac 11271.
        verbose("We can not conclude if the representation is surjective or not. Increasing the parameter A may help.")
        return None

    def non_surjective(self, A=1000):
        r"""
        Return a list of primes p such that the mod-p representation
        *might* not be surjective. If `p` is not in the returned list,
        then the mod-p representation is provably surjective.

        By a theorem of Serre, there are only finitely
        many primes in this list, except when the curve has
        complex multiplication.

        If the curve has CM, we simply return the
        sequence [0] and do no further computation.

        INPUT:

        - ``A`` -- integer (default: 1000); by increasing this parameter
          the resulting set might get smaller

        OUTPUT:

        - ``list`` -- if the curve has CM, returns ``[0]``.
          Otherwise, returns a list of primes where mod-p representation is
          very likely not surjective. At any prime not in this list, the
          representation is definitely surjective.

        EXAMPLES::

            sage: E = EllipticCurve([0, 0, 1, -38, 90])  # 361A
            sage: E.galois_representation().non_surjective()   # CM curve
            [0]

        ::

            sage: E = EllipticCurve([0, -1, 1, 0, 0]) # X_1(11)
            sage: E.galois_representation().non_surjective()
            [5]

            sage: E = EllipticCurve([0, 0, 1, -1, 0]) # 37A
            sage: E.galois_representation().non_surjective()
            []

            sage: E = EllipticCurve([0,-1,1,-2,-1])   # 141C
            sage: E.galois_representation().non_surjective()
            [13]

        ::

            sage: E = EllipticCurve([1,-1,1,-9965,385220]) # 9999a1
            sage: rho = E.galois_representation()
            sage: rho.non_surjective()
            [2]

            sage: E = EllipticCurve('324b1')
            sage: rho = E.galois_representation()
            sage: rho.non_surjective()
            [3, 5]

        ALGORITHM:
        We first find an upper bound `B` on the possible primes. If `E`
        is semi-stable, we can take `B=11` by a result of Mazur. There is
        a bound by Serre in the case that the `j`-invariant is not integral
        in terms of the smallest prime of good reduction. Finally
        there is an unconditional bound by Cojocaru, but which depends
        on the conductor of `E`.
        For the prime below that bound we call ``is_surjective``.
        """
        if self._E.has_cm():
            verbose("cm curve")
            return [0]
        N = self._E.conductor()
        if self._E.is_semistable():
            # Mazur's bound
            C = 11
            verbose("semistable -- so bound is 11")
        elif not self._E.j_invariant().is_integral():
            # prop 24 in Serre
            vs = self._E.j_invariant().denominator().prime_factors()
            C1 = arith.gcd([-arith.valuation(self._E.j_invariant(),v) for v in vs])
            p0 = 2
            while self._E.has_bad_reduction(p0):
                p0 = arith.next_prime(p0+1)
            C2 = (sqrt(p0)+1)**8
            C = max(C1,C2)
            verbose("j is not integral -- Serre's bound is %s" % C)
            C3 = 1 + 4*sqrt(6)*int(N)/3 * sqrt(mul([1+1.0/int(p) for p,_ in arith.factor(N)]))
            C = min(C,C3)
            verbose("conductor = %s, and bound is %s" % (N,C))
        else:
            # Cojocaru's bound (depends on the conductor)
            C = 1 + 4*sqrt(6)*int(N)/3 * sqrt(mul([1+1.0/int(p) for p,_ in arith.factor(N)]))
            verbose("conductor = %s, and bound is %s" % (N,C))
        B = []
        p = 2
        while p <= C:
            t = self.is_surjective(p, A=A)
            verbose("(%s,%s)" % (p,t))
            # both False and None will be appended here.
            if not t:
                B.append(p)
            p = arith.next_prime(p)
        return B

    def image_type(self, p):
        r"""
        Return a string describing the image of the
        mod-p representation.
        The result is provably correct, but only
        indicates what sort of an image we have. If
        one wishes to determine the exact group one
        needs to work a bit harder. The probabilistic
        method of image_classes or Sutherland's galrep
        package can give a very good guess what the
        image should be.

        INPUT:

        - ``p`` -- a prime number

        OUTPUT: string

        EXAMPLES::

            sage: E = EllipticCurve('14a1')
            sage: rho = E.galois_representation()
            sage: rho.image_type(5)
            'The image is all of GL_2(F_5).'

            sage: E = EllipticCurve('11a1')
            sage: rho = E.galois_representation()
            sage: rho.image_type(5)
            'The image is meta-cyclic inside a Borel subgroup as there is a 5-torsion point on the curve.'

            sage: EllipticCurve('27a1').galois_representation().image_type(5)
            'The image is contained in the normalizer of a non-split Cartan group. (cm)'
            sage: EllipticCurve('30a1').galois_representation().image_type(5)
            'The image is all of GL_2(F_5).'
            sage: EllipticCurve("324b1").galois_representation().image_type(5)
            'The image in PGL_2(F_5) is the exceptional group S_4.'

            sage: E = EllipticCurve([0,0,0,-56,4848])
            sage: rho = E.galois_representation()

            sage: rho.image_type(5)
            'The image is contained in the normalizer of a split Cartan group.'

            sage: EllipticCurve('49a1').galois_representation().image_type(7)
            'The image is contained in a Borel subgroup as there is a 7-isogeny.'

            sage: EllipticCurve('121c1').galois_representation().image_type(11)
            'The image is contained in a Borel subgroup as there is a 11-isogeny.'
            sage: EllipticCurve('121d1').galois_representation().image_type(11)
            'The image is all of GL_2(F_11).'
            sage: EllipticCurve('441f1').galois_representation().image_type(13)
            'The image is contained in a Borel subgroup as there is a 13-isogeny.'

            sage: EllipticCurve([1,-1,1,-5,2]).galois_representation().image_type(5)
            'The image is contained in the normalizer of a non-split Cartan group.'
            sage: EllipticCurve([0,0,1,-25650,1570826]).galois_representation().image_type(5)
            'The image is contained in the normalizer of a split Cartan group.'
            sage: EllipticCurve([1,-1,1,-2680,-50053]).galois_representation().image_type(7)    # the dots (...) in the output fix #11937 (installed 'Kash' may give additional output); long time (2s on sage.math, 2014)
            'The image is a... group of order 18.'
            sage: EllipticCurve([1,-1,0,-107,-379]).galois_representation().image_type(7)       # the dots (...) in the output fix #11937 (installed 'Kash' may give additional output); long time (1s on sage.math, 2014)
            'The image is a... group of order 36.'
            sage: EllipticCurve([0,0,1,2580,549326]).galois_representation().image_type(7)
            'The image is contained in the normalizer of a split Cartan group.'

        Test :issue:`14577`::

            sage: EllipticCurve([0, 1, 0, -4788, 109188]).galois_representation().image_type(13)
            'The image in PGL_2(F_13) is the exceptional group S_4.'

        Test :issue:`14752`::

            sage: EllipticCurve([0, 0, 0, -1129345880,-86028258620304]).galois_representation().image_type(11)
            'The image is contained in the normalizer of a non-split Cartan group.'

        For `p=2`::

            sage: E = EllipticCurve('11a1')
            sage: rho = E.galois_representation()
            sage: rho.image_type(2)
            'The image is all of GL_2(F_2), i.e. a symmetric group of order 6.'

            sage: rho = EllipticCurve('14a1').galois_representation()
            sage: rho.image_type(2)
            'The image is cyclic of order 2 as there is exactly one rational 2-torsion point.'

            sage: rho = EllipticCurve('15a1').galois_representation()
            sage: rho.image_type(2)
            'The image is trivial as all 2-torsion points are rational.'

            sage: rho = EllipticCurve('196a1').galois_representation()
            sage: rho.image_type(2)
            'The image is cyclic of order 3.'

        `p=3`::

            sage: rho = EllipticCurve('33a1').galois_representation()
            sage: rho.image_type(3)
            'The image is all of GL_2(F_3).'

            sage: rho = EllipticCurve('30a1').galois_representation()
            sage: rho.image_type(3)
            'The image is meta-cyclic inside a Borel subgroup as there is a 3-torsion point on the curve.'

            sage: rho = EllipticCurve('50b1').galois_representation()
            sage: rho.image_type(3)
            'The image is contained in a Borel subgroup as there is a 3-isogeny.'

            sage: rho = EllipticCurve('3840h1').galois_representation()
            sage: rho.image_type(3)
            'The image is contained in a dihedral group of order 8.'

            sage: rho = EllipticCurve('32a1').galois_representation()
            sage: rho.image_type(3)
            'The image is a semi-dihedral group of order 16, gap.SmallGroup([16,8]).'

        ALGORITHM: Mainly based on Serre's paper.
        """
        if not arith.is_prime(p):
            raise TypeError("p (=%s) must be prime." % p)
        try:
            return self.__image_type[p]
        except KeyError:
            pass
        except AttributeError:
            self.__image_type = {}

        # we eliminate step by step possibilities.
        # checks if the image is all of GL_2, while doing so, it may detect certain other classes already
        # see _is_surjective. It will have set __image_type

        self._is_surjective(p, A=1000)  # this sets __image_type
        try:
            return self.__image_type[p]
        except KeyError:
            pass

        # check if the rep is reducible

        if self.is_reducible(p):
            self.__image_type[p] = "The image is contained in a Borel subgroup as there is a %s-isogeny." % p
            return self.__image_type[p]

        # if we are then the image of rho is not surjective and not contained in a Borel subgroup
        # there are three cases left:
        # it could be in a normalizer of a split Cartan,
        #                  normalizer of a non-split Cartan,
        # or the image in PGL_2 is one of the three exceptional groups A_4 S_4 A_5

        non_split_str = "The image is contained in the normalizer of a non-split Cartan group."
        split_str = "The image is contained in the normalizer of a split Cartan group."
        s4_str = "The image in PGL_2(F_%s) is the exceptional group S_4." % p
        a4_str = "The image in PGL_2(F_%s) is the exceptional group A_4." % p
        a5_str = "The image in PGL_2(F_%s) is the exceptional group A_5." % p

        # we first treat p=3 and 5 separately. p=2 has already been done.

        if p == 3:
            # this implies that the image of rhobar in PGL_2 = S_4
            # determines completely the image of rho
            f = self._E.division_polynomial(3)
            if not f.is_irreducible():
                # must be a product of two polynomials of degree 2
                self.__image_type[p] = "The image is contained in a dihedral group of order 8."
                return self.__image_type[p]
            n = pari(f).polgalois()[0]
            # the following is due to a simple classification of all subgroups of GL_2(F_3)
            if n == 2:
                self.__image_type[p] = "The image is a cyclic group of order 4."
            elif n == 4:
                for ell in prime_range(5,1000):
                    if ell % 3 == 2 and self._E.ap(ell) % 3 != 0:
                        # there is an element of order 8 in the image
                        self.__image_type[p] = "The image is a cyclic group of order 8."
                        return self.__image_type[p]
                self.__image_type[p] = "The image is a group of order 8, most likely a quaternion group."
            elif n == 6:
                self.__image_type[p] = "The image is a dihedral group of order 12."
            elif n == 8:
                self.__image_type[p] = "The image is a semi-dihedral group of order 16, gap.SmallGroup([16,8])."
            elif n == 12:
                self.__image_type[p] = "The image is SL_2(F_3)."
            else:
                raise RuntimeError("Bug in image_type for p = 3.")
            return self.__image_type[p]

        # we also eliminate cm curves

        if self._E.has_cm():
            if self._E.is_good(p) and self._E.is_ordinary(p):
                self.__image_type[p] = split_str + " (cm)"
                return self.__image_type[p]
            if self._E.is_supersingular(p):
                self.__image_type[p] = non_split_str + " (cm)"
                return self.__image_type[p]
            # if the reduction is bad (additive nec.) then
            # the image is in a Borel subgroup and we should have found this before
            raise NotImplementedError("image_type is not implemented for cm-curves at bad prime.")

        # now to p=5 where a lot of non-standard thing happen
        # we run through primes if we hit an element of order 6 in PGL_2, we know that it is the normaliser of a NON-split Cartan
        # if we find both an element of order 3 and one of order 4, we know that we have a S_4 in PGL_2

        if p == 5:
            # we filter here a few cases and leave the rest to the computation of the Galois group later
            ell = 1
            k = GF(p)
            Np = self._E.conductor()*p
            has_an_el_order_4 = False
            has_an_el_order_3 = False
            while ell < 10000:
                ell = arith.next_prime(ell)
                if Np % ell != 0:
                    a_ell = self._E.ap(ell)
                    u = k(a_ell)**2 * k(ell)**(-1)
                    if u == 3:
                        verbose("found an element of order 6", level=2)
                        # found an element of order 6:
                        self.__image_type[p] = non_split_str
                        return self.__image_type[p]

                    if u == 2 and not has_an_el_order_4:
                        # found an element of order 4
                        verbose("found an element of order 4", level=2)
                        has_an_el_order_4 = True
                        if has_an_el_order_3:
                            self.__image_type[p] = s4_str
                            return self.__image_type[p]

                    if u == 1 and not has_an_el_order_3:
                        # found an element of order 3
                        verbose("found an element of order 3", level=2)
                        has_an_el_order_3 = True
                        if has_an_el_order_4:
                            self.__image_type[p] = s4_str
                            return self.__image_type[p]

            verbose("p=5 and we could not determine the image, yet", level=2)
            # we have not yet determined the image, there are only the following possible subgroups of PGL_2
            # (unless we were unlucky and none of the elements of order 6 showed up above, for instance)
            # A_4       of order 12 with elements of order 2 and 3
            # D_4       of order  8 with elements of order 2 and 4  (normalizer of the SPLIT)
            # Z/2 x Z/2  of order 4 with elements of order 2        (in the normalizer of the SPLIT)
            # S_3        of order 6 with elements of order 2 and 3  (inside the normalizer of the NON-split)

            # we compute the splitting field of the 5-division polynomial. Its degree is equal to the above order or the double of it.
            # That allows us to determine almost all cases.

            f = self._E.division_polynomial(5)
            from sage.rings.number_field.splitting_field import SplittingFieldAbort
            try:
                K = f.splitting_field('x', degree_multiple=240, abort_degree=24)
            except SplittingFieldAbort:
                pass
            else:
                if K.degree() in (4,8,16):
                    self.__image_type[p] = split_str
                    return self.__image_type[p]
                if K.degree() == 24:
                    self.__image_type[p] = a4_str
                    return self.__image_type[p]
                if K.degree() == 6:
                    self.__image_type[p] = non_split_str
                    return self.__image_type[p]
                if K.degree() == 12:
                    # PGL - image could be a S_3 in the normalizer of the split or A4
                    self.__image_type[p] = "The image is of order 6 or 12. Probably contained in the normalizer of the split Cartan g."
                    return self.__image_type[p]

        ## now E has no cm, is not semi-stable,
        ## p > 5,
        ## rho_p it is (most probably) not surjective , it is not contained in a Borel subgroup.
        # trying to detect that the image is not exceptional in PGL_2
        # this uses Serre 2.6.iii and Prop 19
        # the existence of certain classes in the image rules out certain cases.
        # we run over the small prime until we are left with only one case
        # for p = 5 this could never distinguish any from an exceptional S_4 or A_4,
        # that is why the case 5 is treated a part before

        else:
            ex_setp = _ex_set(p)
            ell = 1
            k = GF(p)
            Np = self._E.conductor()*p
            could_be_exc = 1
            could_be_split = 1
            could_be_non_split = 1
            # loops over primes as long as we still have two options left
            while ell < 10000 and (could_be_exc + could_be_split + could_be_non_split > 1):
                ell = arith.next_prime(ell)
                if Np % ell != 0:
                    a_ell = self._E.ap(ell)
                    u = k(a_ell)**2 * k(ell)**(-1)
                    if (u not in ex_setp) and could_be_exc == 1:
                        # it can not be in the exceptional
                        verbose("the image cannot be exceptional, found u=%s" % u, level=2)
                        could_be_exc = 0
                    if a_ell != 0 and arith.kronecker(a_ell**2 - 4*ell,p) == 1 and could_be_non_split == 1:
                        # it can not be in the normalizer of the non-split Cartan
                        verbose("the image cannot be non-split, found u=%s" % u, level=2)
                        could_be_non_split = 0
                    if a_ell != 0 and arith.kronecker(a_ell**2 - 4*ell,p) == -1 and could_be_split == 1:
                        # it can not be in the normalizer of the split Cartan
                        verbose("the image cannot be split, found u=%s" % u, level=2)
                        could_be_split = 0

            assert could_be_exc + could_be_split + could_be_non_split > 0, "bug in image_type."

            if could_be_exc + could_be_split + could_be_non_split == 1:
                # it is only one of the three cases:
                if could_be_split == 1:
                    self.__image_type[p] = split_str
                    return self.__image_type[p]
                if could_be_non_split == 1:
                    self.__image_type[p] = non_split_str
                    return self.__image_type[p]
                if could_be_exc == 1:
                    # here we can distinguish further
                    could_be_a4 = 1
                    could_be_s4 = 1
                    could_be_a5 = 1
                    if p % 5 != 1 and p % 5 != 4:
                        could_be_a5 = 0
                    # elements of order 5 # bug corrected see trac 14577
                    R = k['X']
                    f = R([1,-3,1]) #(X**2 - 3*X+1)
                    el5 = f.roots()
                    # loops over primes as long as we still have two options left
                    while ell < 10000 and (could_be_s4 + could_be_a4 + could_be_a5 > 1):
                        ell = arith.next_prime(ell)
                        if Np % ell != 0:
                            a_ell = self._E.ap(ell)
                            u = k(a_ell)**2 * k(ell)**(-1)
                            if u == 2:
                                # it can not be A4 not A5 as they have no elements of order 4
                                could_be_a4 = 0
                                could_be_a5 = 0
                            if u in el5:
                                # it can not be A4 or S4 as they have no elements of order 5
                                could_be_a4 = 0
                                could_be_s4 = 0

                    assert (could_be_s4 + could_be_a4 + could_be_a5 > 0), "bug in image_type."

                    if could_be_s4 + could_be_a4 + could_be_a5 == 1:
                        if could_be_s4 == 1:
                            self.__image_type[p] = s4_str
                            return self.__image_type[p]
                        if could_be_a4 == 1:
                            self.__image_type[p] = a4_str
                            return self.__image_type[p]
                        if could_be_a5 == 1:
                            self.__image_type[p] = a5_str
                            return self.__image_type[p]

                    else:
                        self.__image_type[p] = "The image in PGL_2(F_%s) is an exceptional group A_4, S_4 or A_5, but we could not determine which one." % p
                        return self.__image_type[p]

        # If all fails, we probably have a fairly small group and we can try to detect it using the galois_group
        if p <= 13:
            K = self._E.division_field(p, 'z')
            d = K.absolute_degree()

            verbose("field of degree %s.  try to compute Galois group" % (d), level=2)
            # If the degree is too big, we have no chance at the Galois
            # group.  K.galois_group calls is_galois which used to rely on
            # pari's Galois group computations, so degree < 12
            try:
                if d > 15:
                    raise Exception()
                G = K.galois_group()
            except Exception:
                self.__image_type[p] = "The image is a group of order %s." % d
                return self.__image_type[p]

            else:
                if G.is_abelian():
                    ab = ""
                else:
                    ab = "non-"
                self.__image_type[p] = "The image is a " + ab + "abelian group of order %s." % G.order()
                return self.__image_type[p]

        ## everything failed :

        self.__image_type[p] = "The image could not be determined. Sorry."
        return self.__image_type[p]

    def image_classes(self, p, bound=10000):
        r"""
        This function returns, given the representation `\rho`
        a list of `p` values that add up to 1, representing the
        frequency of the conjugacy classes of the projective image
        of `\rho` in `PGL_2(\GF{p})`.

        Let `M` be a matrix in `GL_2(\GF{p})`, then define
        `u(M) = \text{tr}(M)^2/\det(M)`, which only depends on the
        conjugacy class of `M` in `PGL_2(\GF{p})`. Hence this defines
        a map `u: PGL_2(\GF{p}) \to \GF{p}`, which is almost
        a bijection between conjugacy classes of the source
        and `\GF{p}` (the elements of order `p` and the identity
        map to `4` and both classes of elements of order 2 map to 0).

        This function returns the frequency with which the values of
        `u` appeared among the images of the Frobenius elements
        `a_{\ell}`at `\ell` for good primes `\ell\neq p` below a
        given ``bound``.

        INPUT:

        - ``p`` -- a prime

        - ``bound`` -- a natural number (default: 10000)

        OUTPUT: list of `p` real numbers in the interval `[0,1]` adding up to 1

        EXAMPLES::

            sage: E = EllipticCurve('14a1')
            sage: rho = E.galois_representation()
            sage: rho.image_classes(5)
            [0.2095, 0.1516, 0.2445, 0.1728, 0.2217]

            sage: E = EllipticCurve('11a1')
            sage: rho = E.galois_representation()
            sage: rho.image_classes(5)
            [0.2467, 0.0000, 0.5049, 0.0000, 0.2484]

        ::

            sage: EllipticCurve('27a1').galois_representation().image_classes(5)
            [0.5839, 0.1645, 0.0000, 0.1702, 0.08143]
            sage: EllipticCurve('30a1').galois_representation().image_classes(5)
            [0.1956, 0.1801, 0.2543, 0.1728, 0.1972]
            sage: EllipticCurve('32a1').galois_representation().image_classes(5)
            [0.6319, 0.0000, 0.2492, 0.0000, 0.1189]
            sage: EllipticCurve('900a1').galois_representation().image_classes(5)
            [0.5852, 0.1679, 0.0000, 0.1687, 0.07824]
            sage: EllipticCurve('441a1').galois_representation().image_classes(5)
            [0.5860, 0.1646, 0.0000, 0.1679, 0.08150]
            sage: EllipticCurve('648a1').galois_representation().image_classes(5)
            [0.3945, 0.3293, 0.2388, 0.0000, 0.03749]

        ::

            sage: EllipticCurve('784h1').galois_representation().image_classes(7)
            [0.5049, 0.0000, 0.0000, 0.0000, 0.4951, 0.0000, 0.0000]
            sage: EllipticCurve('49a1').galois_representation().image_classes(7)
            [0.5045, 0.0000, 0.0000, 0.0000, 0.4955, 0.0000, 0.0000]

            sage: EllipticCurve('121c1').galois_representation().image_classes(11)
            [0.1001, 0.0000, 0.0000, 0.0000, 0.1017, 0.1953, 0.1993, 0.0000, 0.0000, 0.2010, 0.2026]
            sage: EllipticCurve('121d1').galois_representation().image_classes(11)
            [0.08869, 0.07974, 0.08706, 0.08137, 0.1001, 0.09439, 0.09764, 0.08218, 0.08625, 0.1017, 0.1009]

            sage: EllipticCurve('441f1').galois_representation().image_classes(13)
            [0.08232, 0.1663, 0.1663, 0.1663, 0.08232, 0.0000, 0.1549, 0.0000, 0.0000, 0.0000, 0.0000, 0.1817, 0.0000]

        REMARKS:

        Conjugacy classes of subgroups of `PGL_2(\GF{5})`

        For the case `p=5`, the order of an element determines almost the value of `u`:

        +-------+---+---+---+---+--------+
        |`u`    | 0 | 1 | 2 | 3 | 4      |
        +-------+---+---+---+---+--------+
        |orders | 2 | 3 | 4 | 6 | 1 or 5 |
        +-------+---+---+---+---+--------+

        Here we give here the full table of all conjugacy classes of subgroups with the values
        that ``image_classes`` should give (as ``bound`` tends to `\infty`). Comparing with the output
        of the above examples, it is now easy to guess what the image is.

        +---------+-----+------------------------------------------+
        |subgroup |order| frequencies of values of `u`             |
        +=========+=====+==========================================+
        | trivial | 1   | [0.0000, 0.0000, 0.0000, 0.0000, 1.000]  |
        +---------+-----+------------------------------------------+
        | cyclic  | 2   | [0.5000, 0.0000, 0.0000, 0.0000, 0.5000] |
        +---------+-----+------------------------------------------+
        | cyclic  |  2  |  [0.5000, 0.0000, 0.0000, 0.0000, 0.5000]|
        +---------+-----+------------------------------------------+
        | cyclic  |  3  |  [0.0000, 0.6667, 0.0000, 0.0000, 0.3333]|
        +---------+-----+------------------------------------------+
        | Klein   |  4  |  [0.7500, 0.0000, 0.0000, 0.0000, 0.2500]|
        +---------+-----+------------------------------------------+
        | cyclic  |  4  |  [0.2500, 0.0000, 0.5000, 0.0000, 0.2500]|
        +---------+-----+------------------------------------------+
        | Klein   |  4  |  [0.7500, 0.0000, 0.0000, 0.0000, 0.2500]|
        +---------+-----+------------------------------------------+
        | cyclic  |  5  |   [0.0000, 0.0000, 0.0000, 0.0000, 1.000]|
        +---------+-----+------------------------------------------+
        | cyclic  |  6  |  [0.1667, 0.3333, 0.0000, 0.3333, 0.1667]|
        +---------+-----+------------------------------------------+
        | `S_3`   |  6  |  [0.5000, 0.3333, 0.0000, 0.0000, 0.1667]|
        +---------+-----+------------------------------------------+
        | `S_3`   |  6  | [0.5000, 0.3333, 0.0000, 0.0000, 0.1667] |
        +---------+-----+------------------------------------------+
        | `D_4`   |  8  |  [0.6250, 0.0000, 0.2500, 0.0000, 0.1250]|
        +---------+-----+------------------------------------------+
        | `D_5`   |  10 |  [0.5000, 0.0000, 0.0000, 0.0000, 0.5000]|
        +---------+-----+------------------------------------------+
        | `A_4`   |  12 | [0.2500, 0.6667, 0.0000, 0.0000, 0.08333]|
        +---------+-----+------------------------------------------+
        | `D_6`   |  12 | [0.5833, 0.1667, 0.0000, 0.1667, 0.08333]|
        +---------+-----+------------------------------------------+
        | Borel   |  20 |  [0.2500, 0.0000, 0.5000, 0.0000, 0.2500]|
        +---------+-----+------------------------------------------+
        | `S_4`   |  24 | [0.3750, 0.3333, 0.2500, 0.0000, 0.04167]|
        +---------+-----+------------------------------------------+
        | `PSL_2` |  60 |  [0.2500, 0.3333, 0.0000, 0.0000, 0.4167]|
        +---------+-----+------------------------------------------+
        | `PGL_2` |  120| [0.2083, 0.1667, 0.2500, 0.1667, 0.2083] |
        +---------+-----+------------------------------------------+
        """
        res = [0 for i in range(p)]
        co = 0
        ell = 1
        while ell <= bound:
            ell = arith.next_prime(ell)
            if ell != p and self._E.is_good(ell):
                d = (self._E.ap(ell)**2 * ell.inverse_mod(p)) % p
                res[d] += 1
                co += 1
        Rt = RealField(16)
        res = [Rt(x)/Rt(co) for x in res]
        return res

#####################################################################
# classification of ell and p-adic reps
#####################################################################

# ell-adic reps

    def is_unramified(self, p, ell):
        r"""
        Return true if the Galois representation to `GL_2(\ZZ_p)` is unramified at `\ell`, i.e.
        if the inertia group at a place above `\ell` in `\text{Gal}(\bar\QQ/\QQ)` has trivial image in
        `GL_2(\ZZ_p)`.

        For a Galois representation attached to an elliptic curve `E`, this
        returns ``True`` if `\ell\neq p` and `E` has good reduction at `\ell`.

        INPUT:

        - ``p`` -- a prime
        - ``ell`` -- another prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('20a3').galois_representation()
            sage: rho.is_unramified(5,7)
            True
            sage: rho.is_unramified(5,5)
            False
            sage: rho.is_unramified(7,5)
            False

        This says that the 5-adic representation is unramified at 7, but the 7-adic representation is ramified at 5.
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        if not arith.is_prime(ell):
            raise ValueError('ell (=%s) must be prime' % ell)
        return (ell != p) and self._E.has_good_reduction(ell)

    def is_unipotent(self, p, ell):
        r"""
        Return true if the Galois representation to `GL_2(\ZZ_p)` is unipotent at `\ell\neq p`, i.e.
        if the inertia group at a place above `\ell` in `\text{Gal}(\bar\QQ/\QQ)` maps into a Borel subgroup.

        For a Galois representation attached to an elliptic curve `E`, this returns ``True`` if
        `E` has semi-stable reduction at `\ell`.

        INPUT:

        - ``p`` -- a prime
        - ``ell`` -- a different prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('120a1').galois_representation()
            sage: rho.is_unipotent(2,5)
            True
            sage: rho.is_unipotent(5,2)
            False
            sage: rho.is_unipotent(5,7)
            True
            sage: rho.is_unipotent(5,3)
            True
            sage: rho.is_unipotent(5,5)
            Traceback (most recent call last):
            ...
            ValueError: unipotent is not defined for l = p, use semistable instead.
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        if not arith.is_prime(ell):
            raise ValueError('ell (=%s) must be prime' % ell)
        if ell == p:
            raise ValueError("unipotent is not defined for l = p, use semistable instead.")
        return not self._E.has_additive_reduction(ell)

    def is_quasi_unipotent(self, p, ell):
        r"""
        Return true if the Galois representation to `GL_2(\ZZ_p)` is quasi-unipotent at `\ell\neq p`, i.e. if there is a finite extension `K/\QQ` such that the inertia group at a place above `\ell` in `\text{Gal}(\bar\QQ/K)` maps into a Borel subgroup.

        For a Galois representation attached to an elliptic curve `E`, this returns always True.

        INPUT:

        - ``p`` -- a prime
        - ``ell`` -- a different prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('11a3').galois_representation()
            sage: rho.is_quasi_unipotent(11,13)
            True
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        if not arith.is_prime(ell):
            raise ValueError('ell (=%s) must be prime' % ell)
        if ell == p:
            raise ValueError("quasi unipotent is not defined for l = p, use semistable instead.")
        return True

# p-adic reps

    def is_ordinary(self, p):
        r"""
        Return true if the `p`-adic Galois representation to `GL_2(\ZZ_p)` is ordinary, i.e.
        if the image of the decomposition group in `\text{Gal}(\bar\QQ/\QQ)` above he prime `p` maps into
        a Borel subgroup.

        For an elliptic curve `E`, this is to ask whether `E` is ordinary at `p`, i.e. good ordinary or multiplicative.

        INPUT:

        - ``p`` -- a prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('11a3').galois_representation()
            sage: rho.is_ordinary(11)
            True
            sage: rho.is_ordinary(5)
            True
            sage: rho.is_ordinary(19)
            False
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        if self._E.has_additive_reduction(p):
            raise NotImplementedError('is_ordinary is only implemented for semi-stable representations')
        return self._E.has_multiplicative_reduction(p) or (self._E.has_good_reduction(p) and self._E.ap(p) % p != 0)

    def is_crystalline(self, p):
        r"""
        Return true is the `p`-adic Galois representation to `GL_2(\ZZ_p)` is crystalline.

        For an elliptic curve `E`, this is to ask whether `E` has good reduction at `p`.

        INPUT:

        - ``p`` -- a prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('64a1').galois_representation()
            sage: rho.is_crystalline(5)
            True
            sage: rho.is_crystalline(2)
            False
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        return self._E.has_good_reduction(p)

    def is_potentially_crystalline(self, p):
        r"""
        Return true is the `p`-adic Galois representation to `GL_2(\ZZ_p)` is potentially crystalline, i.e.
        if there is a finite extension `K/\QQ_p` such that the `p`-adic representation becomes crystalline.

        For an elliptic curve `E`, this is to ask whether `E` has potentially good reduction at `p`.

        INPUT:

        - ``p`` -- a prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('37b1').galois_representation()
            sage: rho.is_potentially_crystalline(37)
            False
            sage: rho.is_potentially_crystalline(7)
            True
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        return self._E.j_invariant().valuation(p) >= 0

    def is_semistable(self, p):
        r"""
        Return true if the `p`-adic Galois representation to `GL_2(\ZZ_p)` is semistable.

        For an elliptic curve `E`, this is to ask whether `E` has semistable reduction at `p`.

        INPUT:

        - ``p`` -- a prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('20a3').galois_representation()
            sage: rho.is_semistable(2)
            False
            sage: rho.is_semistable(3)
            True
            sage: rho.is_semistable(5)
            True
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        return not self._E.has_additive_reduction(p)

    def is_potentially_semistable(self, p):
        r"""
        Return true if the `p`-adic Galois representation to `GL_2(\ZZ_p)` is potentially semistable.

        For an elliptic curve `E`, this returns ``True`` always

        INPUT:

        - ``p`` -- a prime

        OUTPUT: boolean

        EXAMPLES::

            sage: rho = EllipticCurve('27a2').galois_representation()
            sage: rho.is_potentially_semistable(3)
            True
        """
        if not arith.is_prime(p):
            raise ValueError('p (=%s) must be prime' % p)
        return True



from sage.misc.lazy_import import lazy_import

lazy_import('sage.rings.valuation.gauss_valuation', 'GaussValuation')
lazy_import('sage.rings.valuation', 'valuations_catalog', 'valuations')
lazy_import('sage.rings.valuation.value_group', 'DiscreteValueGroup')
del lazy_import



from sage.structure.element cimport Element, ModuleElement, RingElement
from sage.rings.polynomial.polynomial_element cimport Polynomial

from sage.libs.ntl.ntl_ZZ_p cimport ntl_ZZ_p
from sage.libs.ntl.ntl_ZZ_pX cimport ntl_ZZ_pX
from sage.libs.ntl.ntl_ZZ_pContext cimport ntl_ZZ_pContext_class

from sage.libs.ntl.ntl_lzz_p cimport ntl_zz_p
from sage.libs.ntl.ntl_lzz_pX cimport ntl_zz_pX
from sage.libs.ntl.ntl_lzz_pContext cimport ntl_zz_pContext_class

from sage.rings.integer cimport Integer

from sage.libs.ntl.ZZ_pX cimport *
from sage.libs.ntl.lzz_pX cimport *


cdef class Polynomial_dense_mod_n(Polynomial):
    cdef object _poly
    cdef object __singular

cdef class Polynomial_dense_modn_ntl_zz(Polynomial_dense_mod_n):
    cdef zz_pX_c x
    cdef ntl_zz_pContext_class c
    cdef Polynomial_dense_modn_ntl_zz _new(self)
    cpdef _mod_(self, right)

cdef class Polynomial_dense_modn_ntl_ZZ(Polynomial_dense_mod_n):
    cdef ZZ_pX_c x
    cdef ntl_ZZ_pContext_class c
    cdef Polynomial_dense_modn_ntl_ZZ _new(self)
    cpdef _mod_(self, right)

cdef class Polynomial_dense_mod_p(Polynomial_dense_mod_n):
    pass



# sage.doctest: needs sage.combinat sage.modules
r"""
Key and Atom Polynomials

Key polynomials (also known as type A Demazure characters) are defined by
applying the divided difference operator `\pi_{\sigma}`, where `\sigma` is
a permutation, to a monomial corresponding to an integer partition
`\mu \vdash n`.

.. SEEALSO::

    For Demazure characters in other types, see

    - :meth:`sage.combinat.root_system.weyl_characters.WeylCharacterRing.demazure_character`
    - :meth:`sage.categories.classical_crystals.ClassicalCrystals.ParentMethods.demazure_character`

Atom polynomials (also known as Demazure atoms) are defined analogously by
using a deformed divided difference operator `\bar{\pi}_i = \pi_i - 1`.

AUTHORS:

- Trevor K. Karn (2022-08-17): initial version
- Travis Scrimshaw (2025-10-10): expanded to atoms
"""

# ****************************************************************************
#       Copyright (C) 2022 Trevor K. Karn <karnx018 (at) umn.edu>
#                     2025 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from collections.abc import Collection

from sage.categories.graded_algebras_with_basis import GradedAlgebrasWithBasis
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.integer_vector import IntegerVectors
from sage.combinat.permutation import Permutation
from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.infinite_polynomial_ring import InfinitePolynomialRing, InfinitePolynomialRing_sparse
from sage.rings.polynomial.multi_polynomial_ring_base import MPolynomialRing_base
from sage.rings.polynomial.polynomial_ring import PolynomialRing_commutative
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.structure.element import parent


#####################################################################
#  Helper functions


def sorting_word(alpha):
    r"""
    Get a reduced word for the permutation which sorts ``alpha``
    into a partition.

    The result is a list ``l = [i0, i1, i2, ...]`` where each ``ij``
    is a positive integer such that it applies the simple
    transposition `(i_j, i_j+1)`. The transpositions are applied
    starting with ``i0``, then ``i1`` is applied, followed by ``i2``,
    and so on. See :meth:`sage.combinat.permutation.Permutation.reduced_words`
    for the convention used.

    EXAMPLES::

        sage: IV = IntegerVectors()
        sage: from sage.combinat.key_polynomial import sorting_word
        sage: list(sorting_word(IV([2,3,2]))[0])
        [1]
        sage: sorting_word(IV([2,3,2]))[1]
        [3, 2, 2]
        sage: list(sorting_word(IV([5,6,7]))[0])
        [1, 2, 1]
        sage: list(sorting_word(IV([0,3,2]))[0])
        [2, 1]
        sage: list(sorting_word(IV([0,3,0,2]))[0])
        [2, 3, 1]
        sage: list(sorting_word(IV([3,2,1]))[0])
        []
        sage: list(sorting_word(IV([2,3,3]))[0])
        [2, 1]
    """
    w = []
    L = list(alpha)
    n = len(L)

    # bubble sort to get the shortest sorting word
    for i in range(n-1):
        for j in range(n-i-1):
            if L[j] < L[j + 1]:
                w.append(j+1)
                L[j], L[j + 1] = L[j + 1], L[j]
    return reversed(w), L


#####################################################################
#  Divided difference operators


def divided_difference(f, i):
    r"""
    Apply the ``i``-th divided difference operator to the polynomial ``f``.

    EXAMPLES::

        sage: from sage.combinat.key_polynomial import divided_difference
        sage: k = KeyPolynomials(QQ)
        sage: z = k.poly_gens()
        sage: f = z[1]*z[2]^3 + z[1]*z[2]*z[3]
        sage: divided_difference(f, 3)
        z_3^2*z_1 + z_3*z_2*z_1 + z_2^2*z_1

        sage: k = KeyPolynomials(QQ, 4)
        sage: z = k.poly_gens()
        sage: f = z[1]*z[2]^3 + z[1]*z[2]*z[3]
        sage: divided_difference(f, 3)
        z_1*z_2^2 + z_1*z_2*z_3 + z_1*z_3^2

        sage: k = KeyPolynomials(QQ)
        sage: R = k.polynomial_ring(); R
        Infinite polynomial ring in z over Rational Field
        sage: z = R.gen()
        sage: divided_difference(z[1]*z[2]^3, 2)
        -z_2^2*z_1 - z_2*z_1^2
        sage: divided_difference(z[1]*z[2]*z[3], 3)
        0
        sage: divided_difference(z[1]*z[2]*z[3], 4)
        z_2*z_1
        sage: divided_difference(z[1]*z[2]*z[4], 4)
        -z_2*z_1

        sage: k = KeyPolynomials(QQ, 5)
        sage: z = k.polynomial_ring().gens()
        sage: divided_difference(z[1]*z[2]^3, 2)
        -z_1^2*z_2 - z_1*z_2^2
        sage: divided_difference(z[1]*z[2]*z[3], 3)
        0
        sage: divided_difference(z[1]*z[2]*z[3], 4)
        z_1*z_2
        sage: divided_difference(z[1]*z[2]*z[4], 4)
        -z_1*z_2
    """
    P = parent(f)
    if isinstance(P, InfinitePolynomialRing_sparse):
        z = P.gen()
    else:
        z = P.gens()

    si_f = f.subs({z[i]: z[i-1], z[i-1]: z[i]})
    return (si_f - f) // (z[i] - z[i-1])


def isobaric_divided_difference(f, w):
    r"""
    Apply the isobaric divided difference operator `\pi_w` to the
    polynomial `f`.

    ``w`` may be either a single index or a list of
    indices of simple transpositions.

    .. WARNING::

        The simple transpositions should be applied from left to right.

    EXAMPLES::

        sage: from sage.combinat.key_polynomial import isobaric_divided_difference as idd
        sage: R.<z> = InfinitePolynomialRing(GF(3))
        sage: idd(z[1]^4*z[2]^2*z[4], 4)
        0

        sage: idd(z[1]^4*z[2]^2*z[3]*z[4], 3)
        z_4*z_3^2*z_2*z_1^4 + z_4*z_3*z_2^2*z_1^4

        sage: idd(z[1]^4*z[2]^2*z[3]*z[4], [3, 4])
        z_4^2*z_3*z_2*z_1^4 + z_4*z_3^2*z_2*z_1^4 + z_4*z_3*z_2^2*z_1^4

        sage: idd(z[1]^4*z[2]^2*z[3]*z[4], [4, 3])
        z_4*z_3^2*z_2*z_1^4 + z_4*z_3*z_2^2*z_1^4

        sage: idd(z[1]^2*z[2], [3, 2])
        z_3*z_2^2 + z_3*z_2*z_1 + z_3*z_1^2 + z_2^2*z_1 + z_2*z_1^2
    """
    P = parent(f)
    if isinstance(P, InfinitePolynomialRing_sparse):
        z = P.gen()
    else:
        z = P.gens()

    if not hasattr(w, "__iter__"):  # this allows us to pass i instead of a word
        w = [w]
    for i in w:
        fp = z[i-1] * f
        si_fp = fp.subs({z[i]: z[i-1], z[i-1]: z[i]})
        f = (si_fp - fp) // (z[i] - z[i-1])
    return f


def isobaric_divided_difference_bar(f, w):
    r"""
    Apply the isobaric divided difference operator `\bar{\pi}_w` to the
    polynomial `f`.

    This map is defined by `\bar{\pi}_i := \pi_i - 1`.

    ``w`` may be either a single index or a list of
    indices of simple transpositions.

    .. WARNING::

        The simple transpositions should be applied from left to right.

    EXAMPLES::

        sage: from sage.combinat.key_polynomial import isobaric_divided_difference_bar as iddb
        sage: R.<z> = InfinitePolynomialRing(GF(3))
        sage: iddb(z[1]^4*z[2]^2*z[4], 4)
        -z_4*z_2^2*z_1^4

        sage: iddb(z[1]^4*z[2]^2*z[3]*z[4], 3)
        z_4*z_3^2*z_2*z_1^4

        sage: iddb(z[1]^4*z[2]^2*z[3]*z[4], [3, 4])
        z_4^2*z_3*z_2*z_1^4

        sage: iddb(z[1]^4*z[2]^2*z[3]*z[4], [4, 3])
        0

        sage: iddb(z[1]^2*z[2], [3, 2])
        z_3*z_2^2 + z_3*z_2*z_1
    """
    P = parent(f)
    if isinstance(P, InfinitePolynomialRing_sparse):
        z = P.gen()
    else:
        z = P.gens()

    if not hasattr(w, "__iter__"):  # this allows us to pass i instead of a word
        w = [w]
    for i in w:
        sif = f.subs({z[i]: z[i-1], z[i-1]: z[i]})
        f = z[i] * (f - sif) // (z[i-1] - z[i])
    return f


#####################################################################
#  Main classes


class OperatorPolynomial(CombinatorialFreeModule.Element):
    r"""
    Abstract base class for a polynomial defined recursively by
    divided difference type operators.

    Parents should implement the divided difference operator as
    a ``staticmethod`` ``_operator``.
    """
    def _mul_(self, other):
        r"""
        Multiply the elements ``self`` and ``other``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k([4,3,2]) * k([1,1,1])
            k[5, 4, 3]

            sage: k = KeyPolynomials(QQ, 4)
            sage: k([4,3,2,0]) * k([1,1,1,0])
            k[5, 4, 3, 0]
        """
        return self.parent().from_polynomial(self.expand() * other.expand())

    def expand(self):
        r"""
        Return ``self`` written in the monomial basis (i.e., as an element
        in the corresponding polynomial ring).

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: f = k([4,3,2,1])
            sage: f.expand()
            z_3*z_2^2*z_1^3*z_0^4

            sage: f = k([1,2,3])
            sage: f.expand()
            z_2^3*z_1^2*z_0 + z_2^3*z_1*z_0^2 + z_2^2*z_1^3*z_0
             + 2*z_2^2*z_1^2*z_0^2 + z_2^2*z_1*z_0^3 + z_2*z_1^3*z_0^2
             + z_2*z_1^2*z_0^3
        """
        P = self.parent()
        R = P._polynomial_ring
        out = R.zero()
        z = P.poly_gens()

        for m, c in self.monomial_coefficients().items():
            # find the permutation sorting mu into m
            w, mu = sorting_word(m)

            # create the monomial to apply
            monom = R.prod(z[i] ** mi for i, mi in enumerate(mu) if mi)

            out += c * P._operator(monom, w)

        return out

    to_polynomial = expand

    def divided_difference(self, w):
        r"""
        Apply the divided difference operator `\partial_w` to ``self``.

        The convention is to apply from left to right so if
        ``w = [w1, w2, ..., wm]`` then we apply
        `\partial_{w_2 \cdots w_m} \circ \partial_{w_1}`

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k([3,2,1]).divided_difference(2)
            k[3, 1, 1]
            sage: k([3,2,1]).divided_difference([2,3])
            k[3, 1]

            sage: k = KeyPolynomials(QQ, 4)
            sage: k([3,2,1,0]).divided_difference(2)
            k[3, 1, 1, 0]
        """
        if not isinstance(w, Collection):
            w = [w]
        f = self.expand()
        for wi in w:
            f = divided_difference(f, wi)
        return self.parent().from_polynomial(f)

    def isobaric_divided_difference(self, w):
        r"""
        Apply the isobaric divided difference operator `\pi_w` to ``self``.

        The convention is to apply from left to right so if
        ``w = [w1, w2, ..., wm]`` then we apply
        `\pi_{w_2 \cdots w_m} \circ \pi_{w_1}`

        EXAMPLES::

            sage: a = AtomPolynomials(QQ)
            sage: a([3,2,1]).isobaric_divided_difference(2)
            a[3, 1, 2] + a[3, 2, 1]
            sage: a([3,2,1]).isobaric_divided_difference([2, 3])
            a[3, 1, 0, 2] + a[3, 1, 2] + a[3, 2, 0, 1] + a[3, 2, 1]

            sage: a = AtomPolynomials(QQ, 4)
            sage: a([3,2,1,0]).divided_difference(2)
            a[3, 1, 1, 0]
        """
        if not isinstance(w, Collection):
            w = [w]
        f = self.expand()
        for wi in w:
            f = divided_difference(f, wi)
        return self.parent().from_polynomial(f)


class KeyPolynomial(OperatorPolynomial):
    r"""
    A key polynomial.

    Key polynomials are polynomials that form a basis for a polynomial ring
    and are indexed by weak compositions.

    Elements should be created by first creating the basis
    :class:`KeyPolynomialBasis` and passing a list representing the indexing
    composition.

    EXAMPLES::

        sage: k = KeyPolynomials(QQ)
        sage: f = k([4,3,2,1]) + k([1,2,3,4]); f
        k[1, 2, 3, 4] + k[4, 3, 2, 1]
        sage: f in k
        True
    """
    def pi(self, w):
        r"""
        Apply the operator `\pi_w` to ``self``.

        ``w`` may be either a ``Permutation`` or a list of indices of simple
        transpositions (1-based).

        The convention is to apply from left to right so if
        ``w = [w1, w2, ..., wm]`` then we apply
        `\pi_{w_2 \cdots w_m} \circ \pi_{w_1}`

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k([3,2,1]).pi(2)
            k[3, 1, 2]
            sage: k([3,2,1]).pi([2,1])
            k[1, 3, 2]
            sage: k([3,2,1]).pi(Permutation([3,2,1]))
            k[1, 2, 3]
            sage: f = k([3,2,1]) + k([3,2,1,1])
            sage: f.pi(2)
            k[3, 1, 2] + k[3, 1, 2, 1]
            sage: k.one().pi(1)
            k[]

            sage: k([3,2,1,0]).pi(2).pi(2)
            k[3, 1, 2]
            sage: (-k([3,2,1,0]) + 4*k([3,1,2,0])).pi(2)
            3*k[3, 1, 2]

            sage: k = KeyPolynomials(QQ, 4)
            sage: k([3,2,1,0]).pi(2)
            k[3, 1, 2, 0]
            sage: k([3,2,1,0]).pi([2,1])
            k[1, 3, 2, 0]
            sage: k([3,2,1,0]).pi(Permutation([3,2,1,4]))
            k[1, 2, 3, 0]
            sage: f = k([3,2,1,0]) + k([3,2,1,1])
            sage: f.pi(2)
            k[3, 1, 2, 0] + k[3, 1, 2, 1]
            sage: k.one().pi(1)
            k[0, 0, 0, 0]

        TESTS:

        We check that this is consistent with the definition via the
        isobaric divided difference operators::

            sage: from sage.combinat.key_polynomial import isobaric_divided_difference as idd
            sage: k = KeyPolynomials(QQ, 4)
            sage: S4 = Permutations(4)
            sage: f = k([4,2,2,0])
            sage: all(idd(f.expand(), w.reduced_word()) == f.pi(w).expand() for w in S4)
            True

            sage: f = k([4,2,0,1]) - 3 * k([2,0,1,2])
            sage: all(idd(f.expand(), w.reduced_word()) == f.pi(w).expand() for w in S4)
            True
        """
        P = self.parent()
        if isinstance(w, Permutation):
            w = w.reduced_word()
        if not isinstance(w, Collection):
            w = [w]

        if not w or not self:
            return self

        N = max(w) + 1

        if P._k is not None and N > P._k:
            raise ValueError(f"pi_{N-1} does not exist for this polynomial ring")

        ret = P.element_class(P, {})
        for m, c in self._monomial_coefficients.items():
            m = list(m)
            n = len(m)
            for i in w:
                if i > n:
                    continue
                if i == n:
                    m += [0]
                    n += 1
                if m[i-1] <= m[i]:
                    continue
                m[i-1], m[i] = m[i], m[i-1]
            m = P._indices(m)
            if P._k is None:
                m = m.trim()
            if m in ret._monomial_coefficients:
                ret._monomial_coefficients[m] += c
            else:
                ret._monomial_coefficients[m] = c
            if not ret._monomial_coefficients[m]:
                del ret._monomial_coefficients[m]
        return ret

    isobaric_divided_difference = pi

    def pibar(self, w):
        r"""
        Apply the operator `\bar{\pi}_w` to ``self``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k[2,3,1].pibar(2)
            k[2, 1, 3] - k[2, 3, 1]
            sage: k[2,3,1].pibar([2,3])
            k[2, 1, 0, 3] - k[2, 1, 3] - k[2, 3, 0, 1] + k[2, 3, 1]
            sage: k[2,3,1].pibar([2,2]) == -k[2,3,1].pibar(2)
            True

        TESTS:

        We check that this is consistent with the definition via the
        isobaric divided difference operators::

            sage: from sage.combinat.key_polynomial import isobaric_divided_difference_bar as iddb
            sage: k = KeyPolynomials(QQ, 4)
            sage: S4 = Permutations(4)
            sage: f = k[4,2,2,0]
            sage: all(iddb(f.expand(), w.reduced_word()) == f.pibar(w).expand() for w in S4)
            True

            sage: f = k([4,2,0,1]) - 3 * k([2,0,1,2])
            sage: all(iddb(f.expand(), w.reduced_word()) == f.pibar(w).expand() for w in S4)
            True
        """
        if isinstance(w, Permutation):
            w = w.reduced_word()
        if not isinstance(w, Collection):
            w = [w]

        if not w or not self:
            return self

        ret = self
        for i in w:
            ret = ret.pi([i]) - ret
            if not ret:
                return ret
        return ret


class AtomPolynomial(OperatorPolynomial):
    r"""
    An atom polynomial.

    Atom polynomials are polynomials that form a basis for a polynomial
    ring and are indexed by weak compositions.

    Elements should be created by first creating the basis
    :class:`AtomPolynomialBasis` and passing a list representing
    the indexing composition.

    EXAMPLES::

        sage: a = AtomPolynomials(QQ)
        sage: f = a([4,3,2,1]) + a([1,2,3,4]); f
        a[1, 2, 3, 4] + a[4, 3, 2, 1]
        sage: f in a
        True
    """
    def pibar(self, w):
        r"""
        Apply the operator `\bar{\pi}_w` to ``self``.

        ``w`` may be either a ``Permutation`` or a list of indices of simple
        transpositions (1-based).

        The convention is to apply from left to right so if
        ``w = [w1, w2, ..., wm]`` then we apply
        `\bar{\pi}_{w_2 \cdots w_m} \circ \bar{\pi}_{w_1}`.

        EXAMPLES::

            sage: a = AtomPolynomials(QQ)
            sage: a([3,2,1]).pibar(2)
            a[3, 1, 2]
            sage: a([3,2,1]).pibar([2,1])
            a[1, 3, 2]
            sage: a([3,2,1]).pibar(Permutation([3,2,1]))
            a[1, 2, 3]
            sage: f = a([3,2,1]) + a([3,2,1,1])
            sage: f.pibar(2)
            a[3, 1, 2] + a[3, 1, 2, 1]
            sage: a.one().pibar(1)
            0

            sage: a([3,2,1,0]).pibar(2).pibar(2)
            -a[3, 1, 2]
            sage: (-a([3,2,1,0]) + 4*a([3,1,2,0])).pibar(2)
            -5*a[3, 1, 2]

            sage: a = AtomPolynomials(QQ, 4)
            sage: a([3,2,1,0]).pibar(2)
            a[3, 1, 2, 0]
            sage: a([3,2,1,0]).pibar([2,1])
            a[1, 3, 2, 0]
            sage: a([3,2,1,0]).pibar(Permutation([3,2,1,4]))
            a[1, 2, 3, 0]
            sage: f = a([3,2,1,0]) + a([3,2,1,1])
            sage: f.pibar(2)
            a[3, 1, 2, 0] + a[3, 1, 2, 1]
            sage: a.one().pibar(1)
            0

        TESTS:

        We check that this is consistent with the definition via the
        defining isobaric divided difference operators::

            sage: from sage.combinat.key_polynomial import isobaric_divided_difference_bar as iddb
            sage: a = AtomPolynomials(QQ, 4)
            sage: S4 = Permutations(4)
            sage: f = a([4,2,2,0])
            sage: all(iddb(f.expand(), w.reduced_word()) == f.pibar(w).expand() for w in S4)
            True

            sage: f = a([4,2,0,1]) - 3 * a([2,0,1,2])
            sage: all(iddb(f.expand(), w.reduced_word()) == f.pibar(w).expand() for w in S4)
            True
        """
        P = self.parent()
        if isinstance(w, Permutation):
            w = w.reduced_word()
        if not isinstance(w, Collection):
            w = [w]

        if not w or not self:
            return self

        N = max(w) + 1

        if P._k is not None and N > P._k:
            raise ValueError(f"pi_{N-1} does not exist for this polynomial ring")

        ret = P.element_class(P, {})
        for m, c in self._monomial_coefficients.items():
            m = list(m)
            n = len(m)
            sign = 1
            for i in w:
                if i == n:
                    m += [0]
                    n += 1
                if i > n or m[i-1] == m[i]:
                    m = None
                    break
                if m[i-1] < m[i]:
                    sign = -sign
                    continue
                m[i-1], m[i] = m[i], m[i-1]
            if m is None:
                continue
            m = P._indices(m)
            if P._k is None:
                m = m.trim()
            if m in ret._monomial_coefficients:
                ret._monomial_coefficients[m] += c * sign
            else:
                ret._monomial_coefficients[m] = c * sign
            if not ret._monomial_coefficients[m]:
                del ret._monomial_coefficients[m]
        return ret

    def pi(self, w):
        r"""
        Return the action of the isobaric divided difference operator `\pi_w`
        on ``self``.

        EXAMPLES::

            sage: a = AtomPolynomials(QQ)
            sage: a[2,3,1].pi(2)
            a[2, 1, 3] + a[2, 3, 1]
            sage: a[2,3,1].pi([2,3])
            a[2, 1, 0, 3] + a[2, 1, 3] + a[2, 3, 0, 1] + a[2, 3, 1]
            sage: a[2,3,1].pi([2,2]) == a[2,3,1].pi(2)
            True

        TESTS:

        We check that this is consistent with the definition via the
        isobaric divided difference operators::

            sage: from sage.combinat.key_polynomial import isobaric_divided_difference as idd
            sage: a = AtomPolynomials(QQ, 4)
            sage: S4 = Permutations(4)
            sage: f = a[4,2,2,0]
            sage: all(idd(f.expand(), w.reduced_word()) == f.pi(w).expand() for w in S4)
            True

            sage: f = a([4,2,0,1]) - 3 * a([2,0,1,2])
            sage: all(idd(f.expand(), w.reduced_word()) == f.pi(w).expand() for w in S4)
            True
        """
        if isinstance(w, Permutation):
            w = w.reduced_word()
        if not isinstance(w, Collection):
            w = [w]

        if not w or not self:
            return self

        ret = self
        for i in w:
            ret = ret.pibar([i]) + ret
            if not ret:
                return ret
        return ret

    isobaric_divided_difference = pi


class OperatorPolynomialBasis(CombinatorialFreeModule):
    r"""
    A basis for a polynomial ring defined recursively by divided difference
    type operators such that the result is a triangular change of basis with
    the natural monomial basis.
    """
    @staticmethod
    def __classcall__(cls, R=None, k=None, poly_ring=None, poly_coeffs=False):
        r"""
        Normalize input.

        EXAMPLES::

            sage: KeyPolynomials(InfinitePolynomialRing(QQ, ['x', 'y']))
            Traceback (most recent call last):
            ...
            ValueError: polynomial ring has too many generators

            sage: KeyPolynomials(QQ['t0','t1','t2','t3'])
            Key polynomial basis over Rational Field

            sage: KeyPolynomials(QQ['t'])
            Key polynomial basis over Rational Field

            sage: KeyPolynomials(InfinitePolynomialRing(QQ['t'], 'z'))
            Key polynomial basis over Univariate Polynomial Ring in t over Rational Field

            sage: KeyPolynomials(QQ)
            Key polynomial basis over Rational Field

            sage: KeyPolynomials(QQ, 3)
            Key polynomial basis over Rational Field
        """
        poly_type = (PolynomialRing_commutative,
                     MPolynomialRing_base,
                     InfinitePolynomialRing_sparse)

        if isinstance(R, poly_type):
            # if a polynomial ring is provided, we need to determine
            # if it is meant to be self.polynomial_ring() or self.base_ring()
            if isinstance(R, poly_type[0:2]):
                k = R.ngens()
            if isinstance(R, InfinitePolynomialRing_sparse) and R.ngens() > 1:
                raise ValueError("polynomial ring has too many generators")
            if isinstance(R.base_ring(), poly_type[0:2]):
                # if R is of the form K[t_1, ..., t_n][z_*]
                # or K[t_1, ..., t_n][z_1, ..., z_k]
                return super().__classcall__(cls, k=k, poly_ring=R)
            if poly_coeffs:
                # if R is a polynomial ring, but its base ring is not
                # and poly_coeffs is true, then we should interpret
                # R as the base ring
                return super().__classcall__(cls, R=R)
            return super().__classcall__(cls, k=k, poly_ring=R)
        # if R is not a polynomial ring, we know it is self.base_ring()
        return super().__classcall__(cls, R=R, k=k)

    def __init__(self, R=None, k=None, poly_ring=None):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: R = GF(3)['t'].fraction_field()
            sage: k = KeyPolynomials(QQ)
            sage: TestSuite(k).run()
            sage: k = KeyPolynomials(R)
            sage: TestSuite(k).run()

            sage: k = KeyPolynomials(QQ, 4)
            sage: TestSuite(k).run()
            sage: k = KeyPolynomials(R, 4)
            sage: TestSuite(k).run()
        """
        self._k = k

        if self._k is not None:
            def build_index(m):
                return self._indices(m)
        else:
            def build_index(m):
                mc = m.monomial_coefficients()
                v = [0 for _ in range(max(mc, default=-1) + 1)]
                for i, e in mc.items():
                    v[i] = e
                return self._indices(v)

        self._build_index = build_index

        if R is not None:
            if poly_ring:
                raise ValueError("specify only one of base_ring or poly_ring (not both)")
            if k:
                self._polynomial_ring = PolynomialRing(R, 'z_', k)
            else:
                self._polynomial_ring = InfinitePolynomialRing(R, 'z')
        if poly_ring is not None:
            if R is not None:
                raise ValueError("specify only one of base_ring or poly_ring (not both)")
            R = poly_ring.base_ring()
            self._polynomial_ring = poly_ring

        CombinatorialFreeModule.__init__(self, R, IntegerVectors(k=k),
                                         category=GradedAlgebrasWithBasis(R),
                                         prefix=self._prefix, bracket=False)

    def _coerce_map_from_(self, R):
        r"""
        Return the coercion map from ``R`` if it exists.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: m1 = k([3, 2, 4, 0]); m1
            k[3, 2, 4]
            sage: m2 = k(Composition([3, 2, 4])); m2
            k[3, 2, 4]
            sage: m1 == m2
            True

            sage: R = k.polynomial_ring()
            sage: z = R.gen()
            sage: z[0] * k([4, 3, 3, 2])
            k[5, 3, 3, 2]

            sage: X = SchubertPolynomialRing(QQ)
            sage: k(X([4, 3, 2, 1]))
            k[3, 2, 1]
        """
        P = self._polynomial_ring
        if R is P:
            return self.from_polynomial

        phi = P.coerce_map_from(R)
        if phi is not None:
            return self.coerce_map_from(P) * phi
        return None

    def _monomial(self, x):
        r"""
        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k._monomial(k._indices([1, 3, 2, 0, 2]))
            k[1, 3, 2, 0, 2]
            sage: k([3, 2, 3, 4, 0])
            k[3, 2, 3, 4]
            sage: k = KeyPolynomials(QQ, 5)
            sage: k([3, 2, 3, 4, 0])
            k[3, 2, 3, 4, 0]
        """
        if self._k:
            return self._from_dict({x: self.base_ring().one()}, remove_zeros=False)
        return self._from_dict({x.trim(): self.base_ring().one()}, remove_zeros=False)

    def __getitem__(self, c):
        """
        This method implements the abuses of notations ``k[2,1]``,
        ``k[[2,1]]``, etc.

        INPUT:

        - ``c`` -- anything that can represent an index of a basis element

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k[3]
            k[3]
            sage: k[3, 0, 2]
            k[3, 0, 2]
            sage: k[3, 1, 2, 0, 0]
            k[3, 1, 2]

            sage: k = KeyPolynomials(QQ, 4)
            sage: k[3, 0, 1, 0]
            k[3, 0, 1, 0]
            sage: k[3]
            Traceback (most recent call last):
            ...
            ValueError: [3] doesn't satisfy correct constraints
        """
        C = self._indices
        if not isinstance(c, C.element_class):
            if c in ZZ:
                c = C([c])
            else:
                c = C(c)
        return self._monomial(c)

    @cached_method
    def one_basis(self):
        r"""
        Return the basis element indexing the identity.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k.one_basis()
            []

            sage: k = KeyPolynomials(QQ, 4)
            sage: k.one_basis()
            [0, 0, 0, 0]
        """
        if self._k:
            return self._indices([0] * self._k)
        return self._indices([])

    def degree_on_basis(self, alpha):
        """
        Return the degree of the basis element indexed by ``alpha``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k.degree_on_basis([2,1,0,2])
            5

            sage: k = KeyPolynomials(QQ, 5)
            sage: k.degree_on_basis([2,1,0,2,0])
            5
        """
        return ZZ(sum(alpha))

    def polynomial_ring(self):
        r"""
        Return the polynomial ring associated to ``self``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k.polynomial_ring()
            Infinite polynomial ring in z over Rational Field

            sage: k = KeyPolynomials(QQ, 4)
            sage: k.polynomial_ring()
            Multivariate Polynomial Ring in z_0, z_1, z_2, z_3 over Rational Field
        """
        return self._polynomial_ring

    def poly_gens(self):
        r"""
        Return the polynomial generators for the polynomial ring
        associated to ``self``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: k.poly_gens()
            z_*

            sage: k = KeyPolynomials(QQ, 4)
            sage: k.poly_gens()
            (z_0, z_1, z_2, z_3)
        """
        if self._k:
            return self._polynomial_ring.gens()
        return self._polynomial_ring.gen()

    def from_polynomial(self, f):
        r"""
        Expand a polynomial in terms of ``self``.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: z = k.poly_gens(); z
            z_*
            sage: p = z[0]^4*z[1]^2*z[2]*z[3] + z[0]^4*z[1]*z[2]^2*z[3]
            sage: k.from_polynomial(p)
            k[4, 1, 2, 1]

            sage: all(k(c) == k.from_polynomial(k(c).expand()) for c in IntegerVectors(n=5, k=4))
            True

            sage: T = crystals.Tableaux(['A', 4], shape=[4,2,1,1])
            sage: k.from_polynomial(T.demazure_character([2]))
            k[4, 1, 2, 1]
        """
        if f not in self._polynomial_ring:
            try:  # to accept elements of SymbolicRing
                from sage.calculus.var import var
                f = f.substitute([d == var(f'z_{i}')
                                  for i, d in enumerate(f.variables())])
                f = self._polynomial_ring(f)
            except AttributeError:
                raise ValueError(f"f must be an element of {self._polynomial_ring}")

        out = self.zero()
        while f:
            M = f.monomials()[0]
            c = f.monomial_coefficient(M)

            new_term = self._from_dict({self._build_index(*M.exponents()): c})

            temp = new_term.expand()
            f -= temp
            out += new_term

        return out


class KeyPolynomialBasis(OperatorPolynomialBasis):
    r"""
    The key polynomial basis for a polynomial ring.

    For a full definition, see
    `SymmetricFunctions.com <https://www.symmetricfunctions.com/key.htm>`__.
    Key polynomials are indexed by weak compositions with no trailing zeros,
    and `\sigma` is the permutation of shortest length which sorts the
    indexing composition into a partition.

    EXAMPLES:

    Key polynomials are a basis, indexed by (weak) compositions,
    for polynomial rings::

        sage: k = KeyPolynomials(QQ)
        sage: k([3,0,1,2])
        k[3, 0, 1, 2]
        sage: k([3,0,1,2])/2
        1/2*k[3, 0, 1, 2]
        sage: R = k.polynomial_ring(); R
        Infinite polynomial ring in z over Rational Field

        sage: K = KeyPolynomials(GF(5)); K
        Key polynomial basis over Finite Field of size 5
        sage: 2*K([3,0,1,2])
        2*k[3, 0, 1, 2]
        sage: 5*(K([3,0,1,2]) + K([3,1,1]))
        0

    We can expand them in the standard monomial basis::

        sage: k([3,0,1,2]).expand()
        z_3^2*z_2*z_0^3 + z_3^2*z_1*z_0^3 + z_3*z_2^2*z_0^3
         + 2*z_3*z_2*z_1*z_0^3 + z_3*z_1^2*z_0^3 + z_2^2*z_1*z_0^3
         + z_2*z_1^2*z_0^3

        sage: k([0,0,2]).expand()
        z_2^2 + z_2*z_1 + z_2*z_0 + z_1^2 + z_1*z_0 + z_0^2

    If we have a polynomial, we can express it in the key basis::

        sage: z = R.gen()
        sage: k.from_polynomial(z[2]^2*z[1]*z[0])
        k[1, 1, 2] - k[1, 2, 1]

        sage: f = z[3]^2*z[2]*z[0]^3 + z[3]^2*z[1]*z[0]^3 + z[3]*z[2]^2*z[0]^3 + \
        ....: 2*z[3]*z[2]*z[1]*z[0]^3 + z[3]*z[1]^2*z[0]^3 + z[2]^2*z[1]*z[0]^3 + \
        ....: z[2]*z[1]^2*z[0]^3
        sage: k.from_polynomial(f)
        k[3, 0, 1, 2]

    Since the ring of key polynomials may be regarded as a different choice of
    basis for a polynomial ring, it forms an algebra, so we have
    multiplication::

        sage: k([10,5,2])*k([1,1,1])
        k[11, 6, 3]

    We can also multiply by polynomials in the monomial basis::

        sage: k([10,9,1])*z[0]
        k[11, 9, 1]
        sage: z[0] * k([10,9,1])
        k[11, 9, 1]
        sage: k([10,9,1])*(z[0] + z[3])
        k[10, 9, 1, 1] + k[11, 9, 1]

    When the sorting permutation is the longest element, the key polynomial
    agrees with the Schur polynomial::

        sage: s = SymmetricFunctions(QQ).schur()
        sage: k([1,2,3]).expand()
        z_2^3*z_1^2*z_0 + z_2^3*z_1*z_0^2 + z_2^2*z_1^3*z_0
         + 2*z_2^2*z_1^2*z_0^2 + z_2^2*z_1*z_0^3 + z_2*z_1^3*z_0^2
         + z_2*z_1^2*z_0^3
        sage: s[3,2,1].expand(3)
        x0^3*x1^2*x2 + x0^2*x1^3*x2 + x0^3*x1*x2^2 + 2*x0^2*x1^2*x2^2
         + x0*x1^3*x2^2 + x0^2*x1*x2^3 + x0*x1^2*x2^3

    The polynomial expansions can be computed using crystals and expressed in
    terms of the key basis::

        sage: T = crystals.Tableaux(['A',3],shape=[2,1])
        sage: f = T.demazure_character([3,2,1])
        sage: k.from_polynomial(f)
        k[1, 0, 0, 2]

    The default behavior is to work in a polynomial ring with infinitely many
    variables. One can work in a specicfied number of variables::

        sage: k = KeyPolynomials(QQ, 4)
        sage: k([3,0,1,2]).expand()
        z_0^3*z_1^2*z_2 + z_0^3*z_1*z_2^2 + z_0^3*z_1^2*z_3
         + 2*z_0^3*z_1*z_2*z_3 + z_0^3*z_2^2*z_3 + z_0^3*z_1*z_3^2 + z_0^3*z_2*z_3^2

        sage: k([0,0,2,0]).expand()
        z_0^2 + z_0*z_1 + z_1^2 + z_0*z_2  + z_1*z_2 + z_2^2

        sage: k([0,0,2,0]).expand().parent()
        Multivariate Polynomial Ring in z_0, z_1, z_2, z_3 over Rational Field

    If working in a specified number of variables, the length of the indexing
    composition must be the same as the number of variables::

        sage: k([0,0,2])
        Traceback (most recent call last):
         ...
        TypeError: do not know how to make x (= [0, 0, 2]) an element of self
         (=Key polynomial basis over Rational Field)

    One can also work in a specified polynomial ring::

        sage: k = KeyPolynomials(QQ['x0', 'x1', 'x2', 'x3'])
        sage: k([0,2,0,0])
        k[0, 2, 0, 0]
        sage: k([4,0,0,0]).expand()
        x0^4

    If one wishes to use a polynomial ring as coefficients for the key
    polynomials, pass the keyword argument ``poly_coeffs=True``::

        sage: k = KeyPolynomials(QQ['q'], poly_coeffs=True)
        sage: R = k.base_ring(); R
        Univariate Polynomial Ring in q over Rational Field
        sage: R.inject_variables()
        Defining q
        sage: (q^2 + q + 1)*k([0,2,2,0,3,2])
        (q^2+q+1)*k[0, 2, 2, 0, 3, 2]
    """
    Element = KeyPolynomial
    _name = "Key polynomial basis"
    _prefix = 'k'
    _operator = staticmethod(isobaric_divided_difference)

    def _coerce_map_from_(self, R):
        r"""
        Return the coercion map from ``R`` if it exists.

        EXAMPLES::

            sage: k = KeyPolynomials(QQ)
            sage: m1 = k([3, 2, 4, 0]); m1
            k[3, 2, 4]
            sage: m2 = k(Composition([3, 2, 4])); m2
            k[3, 2, 4]
            sage: m1 == m2
            True

            sage: R = k.polynomial_ring()
            sage: z = R.gen()
            sage: z[0] * k([4, 3, 3, 2])
            k[5, 3, 3, 2]

            sage: X = SchubertPolynomialRing(QQ)
            sage: k(X([4, 3, 2, 1]))
            k[3, 2, 1]
        """
        P = self._polynomial_ring
        if R is P:
            return self.from_polynomial

        from sage.combinat.schubert_polynomial import SchubertPolynomialRing_xbasis
        if isinstance(R, SchubertPolynomialRing_xbasis):
            return self.from_schubert_polynomial

        phi = P.coerce_map_from(R)
        if phi is not None:
            return self.coerce_map_from(P) * phi
        return None

    def from_schubert_polynomial(self, x):
        r"""
        Expand a Schubert polynomial in ``self``.

        EXAMPLES::

            sage: k = KeyPolynomials(ZZ)
            sage: X = SchubertPolynomialRing(ZZ)
            sage: f = X([2,1,5,4,3])
            sage: k.from_schubert_polynomial(f)
            k[1, 0, 2, 1] + k[2, 0, 2] + k[3, 0, 0, 1]
            sage: k.from_schubert_polynomial(2)
            2*k[]
            sage: k(f)
            k[1, 0, 2, 1] + k[2, 0, 2] + k[3, 0, 0, 1]

            sage: k = KeyPolynomials(GF(7), 4)
            sage: k.from_schubert_polynomial(f)
            k[1, 0, 2, 1] + k[2, 0, 2, 0] + k[3, 0, 0, 1]

        TESTS::

            sage: k = KeyPolynomials(ZZ)
            sage: k.from_schubert_polynomial(k([3,2]))
            Traceback (most recent call last):
            ...
            ValueError: not a Schubert polynomial

            sage: k = KeyPolynomials(ZZ)
            sage: X = SchubertPolynomialRing(ZZ)
            sage: it = iter(Compositions())
            sage: for _ in range(50):
            ....:     C = next(it)
            ....:     assert k.from_schubert_polynomial(X(k[C])) == k[C], C

            sage: k = KeyPolynomials(ZZ, 4)
            sage: X = SchubertPolynomialRing(ZZ)
            sage: it = iter(k.basis().keys())
            sage: for _ in range(50):
            ....:     C = next(it)
            ....:     assert k.from_schubert_polynomial(X(k[C])) == k[C], C
        """
        if x in self.base_ring():
            return self(x)

        from sage.combinat.schubert_polynomial import SchubertPolynomial_class
        if not isinstance(x, SchubertPolynomial_class):
            raise ValueError('not a Schubert polynomial')

        from sage.combinat.diagram import RotheDiagram
        out = self.zero()
        if self._k is not None:
            def build_elt(wt):
                wt = list(wt)
                wt += [0] * (self._k - len(wt))
                return self[wt]
        else:
            def build_elt(wt):
                return self[wt]

        for m, c in x.monomial_coefficients().items():
            D = RotheDiagram(m)
            a = self.zero()
            for d in D.peelable_tableaux():
                a += build_elt(d.left_key_tableau().weight())
            out += c * a

        return out


class AtomPolynomialBasis(OperatorPolynomialBasis):
    r"""
    The atom polynomial basis for a polynomial ring.

    For a full definition, see
    `SymmetricFunctions.com <https://www.symmetricfunctions.com/key.htm#atom>`__.
    Atom polynomials are indexed by weak compositions `\alpha` with no
    trailing zeros. A basis element is given by

    .. MATH::

        a_{\alpha} = \bar{\pi}_{\sigma} x^{\mu},

    where `\sigma` is the permutation of shortest length which sorts the
    `\alpha` into a partition `\mu`.

    EXAMPLES:

    Atom polynomials are a basis, indexed by (weak) compositions,
    for polynomial rings::

        sage: a = AtomPolynomials(QQ)
        sage: a([3,0,1,2])
        a[3, 0, 1, 2]
        sage: a([3,0,1,2])/2
        1/2*a[3, 0, 1, 2]
        sage: R = a.polynomial_ring(); R
        Infinite polynomial ring in z over Rational Field

        sage: A = AtomPolynomials(GF(5)); A
        Atom polynomial basis over Finite Field of size 5
        sage: 2*A([3,0,1,2])
        2*a[3, 0, 1, 2]
        sage: 5*(A([3,0,1,2]) + A([3,1,1]))
        0

    We can expand them in the standard monomial basis::

        sage: a([2,1,3,2]).expand()
        z_3^2*z_2^3*z_1*z_0^2 + z_3^2*z_2^2*z_1^2*z_0^2

        sage: a[0,2,2].expand()
        z_2^2*z_1^2 + z_2^2*z_1*z_0 + z_2*z_1^2*z_0

    If we have a polynomial, we can express it in the atom basis::

        sage: z = R.gen()
        sage: a.from_polynomial(z[2]^2*z[1]*z[0])
        a[1, 1, 2]
        sage: a.from_polynomial(z[2]^3*z[1]^2*z[0])
        a[1, 2, 3]

        sage: f = (z[3]^2*z[2]*z[0]^3 + z[3]^2*z[1]*z[0]^3 + z[3]*z[2]^2*z[0]^3
        ....:      + 2*z[3]*z[2]*z[1]*z[0]^3 + z[3]*z[1]^2*z[0]^3
        ....:      + z[2]^2*z[1]*z[0]^3 + z[2]*z[1]^2*z[0]^3)
        sage: a.from_polynomial(f)
        a[3, 0, 1, 2] + a[3, 0, 2, 1] + a[3, 1, 0, 2] + a[3, 1, 2]
         + a[3, 2, 0, 1] + a[3, 2, 1]

    Since the ring of atom polynomials may be regarded as a different
    choice of basis for a polynomial ring, it forms an algebra, so
    we have multiplication::

        sage: a[10,5,2] * a[1,1,1]
        a[11, 6, 3]

    We can also multiply by polynomials in the monomial basis::

        sage: a([10,9,1]) * z[0]
        a[11, 9, 1]
        sage: z[0] * a([10,9,1])
        a[11, 9, 1]
        sage: a([10,9,1]) * (z[0] + z[3])
        a[10, 9, 1, 1] + a[11, 9, 1]
    """
    Element = AtomPolynomial
    _name = "Atom polynomial basis"
    _prefix = 'a'
    _operator = staticmethod(isobaric_divided_difference_bar)

    def from_key_polynomial(self, x):
        r"""
        Expand a key polynomial in ``self``.

        EXAMPLES::

            sage: a = AtomPolynomials(ZZ)
            sage: k = KeyPolynomials(ZZ)
            sage: f = k[1,2,3]
            sage: a.from_key_polynomial(f)
            a[1, 2, 3] + a[1, 3, 2] + a[2, 1, 3] + a[2, 3, 1] + a[3, 1, 2] + a[3, 2, 1]
            sage: a.from_key_polynomial(2)
            2*a[]
            sage: a(f)
            a[1, 2, 3] + a[1, 3, 2] + a[2, 1, 3] + a[2, 3, 1] + a[3, 1, 2] + a[3, 2, 1]

            sage: a = AtomPolynomials(GF(7), 4)
            sage: a.from_key_polynomial(f)
            a[1, 2, 3, 0] + a[1, 3, 2, 0] + a[2, 1, 3, 0]
             + a[2, 3, 1, 0] + a[3, 1, 2, 0] + a[3, 2, 1, 0]

        TESTS::

            sage: a = AtomPolynomials(ZZ)
            sage: a.from_key_polynomial(a([3,2]))
            Traceback (most recent call last):
            ...
            ValueError: not a key polynomial

            sage: a = AtomPolynomials(ZZ)
            sage: k = KeyPolynomials(ZZ)
            sage: it = iter(a.basis().keys())
            sage: for _ in range(50):
            ....:     C = next(it)
            ....:     assert a.from_key_polynomial(k.from_polynomial(a[C].expand())) == a[C], C

            sage: a = AtomPolynomials(ZZ, 4)
            sage: k = KeyPolynomials(ZZ, 4)
            sage: it = iter(a.basis().keys())
            sage: for _ in range(50):
            ....:     C = next(it)
            ....:     assert a.from_key_polynomial(k.from_polynomial(a[C].expand())) == a[C], C
        """
        if x in self.base_ring():
            return self(x)

        if not isinstance(x, KeyPolynomial):
            raise ValueError('not a key polynomial')

        from sage.combinat.permutation import Permutations

        def on_basis(m):
            w, mu = sorting_word(m)
            if self._k and len(mu) < self._k:
                mu += [0] * (self._k - len(mu))
            dom = self[mu]
            w = list(w)
            if not w:
                return dom
            sigma = Permutations(max(w)+1).from_reduced_word(w)
            return self.sum(dom.pibar(wp.reduced_word()) for wp in sigma.bruhat_smaller())

        return self.linear_combination((on_basis(m), c) for m, c in x)

    def _coerce_map_from_(self, R):
        r"""
        Return the coercion map from ``R`` if it exists.

        EXAMPLES::

            sage: a = AtomPolynomials(QQ)
            sage: m1 = a([3, 2, 4, 0]); m1
            a[3, 2, 4]
            sage: m2 = a(Composition([3, 2, 4])); m2
            a[3, 2, 4]
            sage: m1 == m2
            True

            sage: R = a.polynomial_ring()
            sage: z = R.gen()
            sage: z[0] * a([4, 3, 3, 2])
            a[5, 3, 3, 2]

            sage: X = SchubertPolynomialRing(QQ)
            sage: a(X([4, 3, 2, 1]))
            a[3, 2, 1]
        """
        P = self._polynomial_ring
        if R is P:
            return self.from_polynomial

        if isinstance(R, KeyPolynomialBasis):
            return self.from_key_polynomial

        from sage.combinat.schubert_polynomial import SchubertPolynomialRing_xbasis
        if isinstance(R, SchubertPolynomialRing_xbasis):
            K = KeyPolynomialBasis(self.base_ring(), self._k, self._polynomial_ring)
            return self._coerce_map_via([K], R)

        phi = P.coerce_map_from(R)
        if phi is not None:
            return self.coerce_map_from(P) * phi
        return None
